"""
`akid` builds another layer of abstraction on top of *Tensor*: *Block*.
Tensor can be taken as the media/formalism signal propagates in digital world,
while Block is the data processing entity that processes inputs and emits
outputs.

Best designs mimic nature. `akid` tries to reproduce how signals in nature
propagates. Information flow can be abstracted as data propagating through
inter-connected blocks, each of which processes inputs and emits outputs. For
example, a vision classification system is a block that takes image inputs and
gives classification results. Everything is a `Block` in `akid`.

The type one block could be as simple as a convonlutional neural network layer
that merely does convolution on the input data and outputs the results; it also
be as complex as an acyclic graph that inter-connects blocks to build a neural
network, or sequentially linked block system that does data augmentation.

Compared with pure symbol computation approach, like the one in tensorflow, a
block is able to contain states associated with this processing unit. Signals
are passed between blocks in form of tensors or list of tensors. Many heavy
lifting has been done in the block (`Block` and its sub-classes),
e.g. pre-condition setup, name scope maintenance, copy functionality for
validation and copy functionality for distributed replicas, setting up and
gathering visualization summaries, centralization of variable allocation,
attaching debugging ops now and then etc.
"""
from __future__ import absolute_import, division, print_function

import abc
import copy
from deprecation import deprecated

import tensorflow as tf

from ..utils import glog as log
from .common import (
    TRAIN_SUMMARY_COLLECTION,
    VALID_SUMMARY_COLLECTION,
    SPARSITY_SUMMARY_SUFFIX,
    FILTER_WEIGHT_COLLECTION,
)
from ..core import initializers
from .. import backend as A
from .interface_blocks import UpdateBlock
import six


class Block(six.with_metaclass(abc.ABCMeta, object)):
    """
    The top level class. Everything should be its sub-class.

    A `Block` holds computational graph. Sub-class should implement `_setup` to
    build the graph. It does not enforce how the computational graph should be
    used, which should be the responsibility of sub classes.

    Any outputs of a block is of type Tensor. Note no builtin tensor type
    exists. Depending on the backend, the Tensor is the type Tensor from the
    backend. Such a requirement is to normalize data type in outputs.

    `Block` supports add arbitrary code after setup by offering
    `post_setup_hook`. All functions added in the hook will be called after
    setup. `self` is passed as an argument to functions in the hook.

    Call `setup` of each block before using it.
    """
    def __init__(self, name=None, debug=False, **kwargs):
        super(Block, self).__init__(**kwargs)

        A.inc_block_count(self.__class__)

        if name is None:
            self.name = self.get_default_name()
        else:
            self.name = name
        self.debug = debug

        # Hooks that are called after setup.
        self.post_setup_hook = []
        self.pre_setup_hook = []

        self.is_setup = None

        # Enable logging to stderr for each individual, so logging is available
        # when not using the Kid.
        log.init(akid_logger=True)

    def get_default_name(self):
        count = A.get_block_count(self.__class__)
        return "{}_{}".format(self.__class__.NAME, count)

    def setup(self):
        with A.variable_scope(self.name):
            self._pre_setup()
            self._setup()
            self._post_setup()

        self.is_setup = True

    def _setup(self):
        pass

    def teardown(self):
        pass

    def _pre_setup(self):
        for f in self.pre_setup_hook:
            f(self)

    def _post_setup(self):
        for f in self.post_setup_hook:
            f(self)

    def log(self, message, debug=False):
        """
        An unified place where logging should be processed. It is planned to
        handle logging level according to situation.

        Currently, it only handles info and debug level message, given the
        purpose of logging is to provide a detailed trace of operations of
        `akid`.

        NOTE: Remember to initialize it before using. See `utils.glog` for
        details.
        """
        if debug:
            log.debug(message)
        else:
            log.info(message)

    def error(self, message):
        log.error(message)

    def warning(self, message):
        log.warning(message)

    def get_copy(self):
        return copy.copy(self)

    def get_clone(self):
        """
        To clone is to create a copy of this block that has an independent copy
        of each component of this block. The relationship between get_clone and
        get_copy is like the that between deepcopy and copy.
        """
        return self.get_copy()


class DataBlock(Block):
    """
    A block that has a property `data` to supply data. Data does not
    necessarily flow in, or are operated in batches in this block.
    """
    @abc.abstractmethod
    def data(self):
        """
        An abstract method to enforce all sub-classes to provide their
        processed data through this interface.
        """
        raise NotImplementedError("Each concrete block needs to implement this"
                                  " method to provide an interface to offer"
                                  " data!")


class FlowBlock(DataBlock):
    """
    Abstract class for naive data flow. It implements naive data flow
    interfaces without any other functionalities.
    """
    def forward(self, *args, **kwargs):
        """
        Args:
            All arguments will be passed to the actual `_forward` function.
        """
        if not self.is_setup:
            self.setup()

        with A.variable_scope(self.name):
            out = self._forward(*args, **kwargs)

        return out

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)


class ProcessingBlock(FlowBlock):
    """
    Abstract class for an arbitrary block that generates output. `Source`,
    `Sensor`, `ProcessingLayer`, `LossLayer` and `ProcessingSystem` etc are all
    sub-classes of this class.

    `ProcessingBlock` builds computational graph that processes data.

    A `ProcessingBlock` should try to implement most of its functionality only
    with what it owns, and ask for communication (which in implementation is to
    provide interfaces) as little as possible.

    Outputs of a block is in form of properties. `ProcessingBlock` has an
    abstract property data which sub-class should implement to provide the
    processed outputs.

    `forward` is the interface for any containers, such as a `Brain` class,
    that hold this block, to call to do build computational graph that normally
    leads to a forward propagation in neural networks to do posterior inference
    (the statement may not hold exactly when the neural network does not learn
    a probability). The concrete forward propagation of sub-classes should be
    implemented in `_forward` method. The return value of `_forward` will be
    returned by `forward`. To add arbitrary code before and after `_forward`,
    use `pre_forward_hook`; to add after, use `post_forward_hook`. All
    arguments passed to `_forward` are also available in functions in the
    hooks.
    """
    _flags = ["do_summary", "summarize_output", "is_mon", "done_first_batch_monitoring_pass"]

    def __init__(self,
                 do_summary=None,
                 bag=None,
                 summarize_output=None,
                 **kwargs):
        """
        Create a layer and name it.

        Args:
            name: str
                The name of the layer. It also serves as the name scope of this
                layer in tensorflow. If `None`, the default name provided by
                Tensorflow will be used.
            do_summary: Boolean
                Whether to do summary on blocks contained. If True, outputs of
                this `ProcessingBlock` would be added to tensorflow summary.
            bag: supposed to be a dict
                A dictionary that holds any further ad hoc information one
                wants to keep in this block, such as a list of filters you want
                to visualize later.
            summarize_output: bool
                Summarize on output, including data, loss etc. If true, a name
                tag will be given to the outputs, consequently, summarizing
                events will be created for tensorboard. This semantics is
                enforced by each individual layer, by implementing properly.
        """
        super(ProcessingBlock, self).__init__(**kwargs)

        self.do_summary = do_summary
        self.log("{} has bag: {}".format(self.name, bag))
        self.bag = bag
        self.summarize_output = summarize_output

        # Hooks
        self.pre_forward_hook = []
        self.post_forward_hook = []

        # Some operations only run at the first forward pass, so set a flag.
        self.done_first_pass = False

        self.done_first_batch_monitoring_pass = False
        # Whether the brain is in the batch monitoring mode.
        self.is_mon = False

    def set_flag(self, flag_name, v):
        if flag_name in self.__class__._flags:
            setattr(self, flag_name, v)
        else:
            raise ValueError("{} does not have flag {}".format(flag_name))

    def set_do_summary_flag(self, v):
        self.do_summary = v

    def summarize_data(self, data, val=False, collections=None):
        """
        Helper function to summarize data with Tensorboard. `data` needs to be
        numerical, e.g., a Tensor, or a numpy array.

        Args:
            val: bool
                Whether to summarize data in val mode as well. If False,
                Summary ops won't be created for data in val mode won.
        """
        if not self.do_summary:
            return

        if not self.done_first_pass\
           or val and self.is_val and not self.done_first_pass_val:
            if collections is None:
                if not val:
                    collections = [TRAIN_SUMMARY_COLLECTION]
                else:
                    collections = [VALID_SUMMARY_COLLECTION]

            self._data_summary(data, collections=collections)

    def forward(self, *args, **kwargs):
        """
        Args:
            All arguments will be passed to the actual `_forward` function.
        """
        if not self.is_setup:
            self.setup()

        with A.variable_scope(self.name):
            self._pre_forward(*args, **kwargs)
            out = self._forward(*args, **kwargs)
            post_out = self._post_forward(*args, **kwargs)
            if not self.done_first_pass:
                self.done_first_pass = True

        if post_out is None:
            return out
        else:
            return post_out

    def _pre_forward(self, *args, **kwargs):
        for f in self.pre_forward_hook:
            f(self, *args, **kwargs)

    def _post_forward(self, *args, **kwargs):
        for f in self.post_forward_hook:
            f(self, *args, **kwargs)

        if self.done_first_pass or not self.do_summary:
            return

        if self.data is not None:
            self._data_summary(self.data, sparsity_summary=True, collections=[TRAIN_SUMMARY_COLLECTION])

    @abc.abstractmethod
    def _forward(self):
        """
        An abstract method that must be overrided.
        """
        raise NotImplementedError('Each sub-layer needs to implement this'
                                  'method to process data!')

    def _data_summary(self, data, sparsity_summary=False, collections=None):
        """
        Helper function to do statistical summary on the bundle of data.
        """
        if type(data) is not list and type(data) is not tuple:
            data = [data]

        for d in data:
            name = A.get_name(d)
            if name:
                if not A.is_numerical(d):
                    return

                self.log("Do tensorboard summary on outputs {} of {}".format(
                    name, self.name))

                shape = A.get_shape(d)
                if shape != 0:
                    dim = len(shape)
                else:
                    dim = 0
                if dim == 0:
                    A.summary.scalar(name, d, collections=collections)
                else:
                    A.summary.histogram(name,
                                        d,
                                        collections=collections)
                    if sparsity_summary:
                        A.summary.scalar(
                            A.append_suffix(name, SPARSITY_SUMMARY_SUFFIX),
                            A.nn.zero_fraction(d,
                                               # The scope removing does
                                               # nothing for tensorflow
                                               # backend, since the name gotten
                                               # is already scope removed.
                                               name=A.remove_scope_from_name(name) \
                                               + '/' \
                                               + SPARSITY_SUMMARY_SUFFIX),
                            collections=collections)


class ValidatableProcessingBlock(ProcessingBlock):
    """
    Since normally we have the two types of data, a.k.a, validation data and
    training data, `ValidatableProcessingBlock` support blocks to operate in
    two modes.
    """
    _flags = copy.copy(ProcessingBlock._flags)
    _flags.append("do_summary_on_val")

    def __init__(self, do_summary_on_val=None, **kwargs):
        """
        Args:
            do_summary_on_val: bool
                Whether to do summary on validation copy.
        """
        super(ValidatableProcessingBlock, self).__init__(**kwargs)

        # A Boolean flag to indicate whether this block is in validation mode.
        self.mode = "train"
        self.done_first_pass_val = False
        self.do_summary_on_val = do_summary_on_val

    @property
    def is_val(self):
        return self.mode == A.Mode.VAL

    def set_do_summary_on_val_flag(self, v):
        self.do_summary_on_val = v

    def forward(self, *args, **kwargs):
        out = super(ValidatableProcessingBlock, self).forward(*args, **kwargs)

        if not self.done_first_pass_val and self.is_val:
            self.done_first_pass_val = True

        return out

    def _pre_setup(self):
        super(ValidatableProcessingBlock, self)._pre_setup()

        if self.is_val:
            A.get_variable_scope().reuse_variables()

    def get_val_copy(self):
        """
        Get a copy for validation.

        Since a processing layer is learned, it has to be taken out for
        evaluation from time to time.
        """
        if A.backend() == A.TF:
            val_copy = self.get_copy()
            val_copy.name = self.name + '_val'
        elif A.backend() == A.TORCH:
            # Since torch is dynamic graph, no need to create a copy for
            # validation.
            val_copy = self
        else:
            raise ValueError("Not supported backend.")

        val_copy.set_val()
        return val_copy

    def set_val(self, val):
        self.mode = A.Mode.VAL if val else A.Mode.TRAIN

    def _post_forward(self, *args, **kwargs):
        for f in self.post_forward_hook:
            f(self, *args, **kwargs)

        # If has done forward for building computational graphs that exist in
        # form of ops, return. But it is possible we still need to build ops
        # for validation block.
        if self.done_first_pass and not self.is_val:
            return

        if not self.do_summary:
            return

        if not self.is_val\
           or (self.is_val\
               and self.do_summary_on_val\
               and not self.done_first_pass_val):

            collection = VALID_SUMMARY_COLLECTION if self.is_val \
                else TRAIN_SUMMARY_COLLECTION

            if self.data is not None:
                self._data_summary(self.data, sparsity_summary=True, collections=[collection])


class ShadowableBlock(ValidatableProcessingBlock):
    """
    A block for creating shadow replicas for parallelism.

    A shadow replica is a replica of the same parameter of the genuine one, but
    be placed on different device for parallelism. An example would be data
    parallelism in multiple GPU setting or distributed setting.
    """
    def __init__(self, **kwargs):
        super(ShadowableBlock, self).__init__(**kwargs)

        # Whether this processing block is a shallow replica.
        self.is_shadow = False

    def setup(self):
        """
        For shadow copies, only broadcast parameters need to be retrieved
        again, so only `_setup` will be called.
        """
        if self.is_shadow:
            with A.variable_scope(self.name):
                self._setup()
        else:
            super(ShadowableBlock, self).setup()

    def set_shadow(self):
        """
        Set this block to a shadow replica.
        """
        self.is_shadow = True

    def get_shadow_copy(self):
        shadow_copy = self.get_copy()
        shadow_copy.set_shadow()
        return shadow_copy

    def log(self, msg, *args, **kwargs):
        """
        Logging method to control the verbosity of the output. It logs all
        logging of shadow replica to debug.
        """
        # It is possible the constructor of `ShadowableBlock` has not been
        # called, in this case we fall back to its super's log.
        if hasattr(self, "is_shadow") and self.is_shadow:
            log.debug(msg)
        else:
            super(ShadowableBlock, self).log(msg, *args, **kwargs)


class GenerativeBlock(ShadowableBlock):
    """
    Block that optionally supports a backward pass.

    It aims to model generative modeling instead of just discriminative
    training. It supports a `backward` that normally does generative
    reconstruction according to the forward propagation (posterior inference)
    results.
    """
    def backward(self, X):
        """
        According to the top-down inference results, reconstruct the input.

        TODO: should call _backward somehow maybe
        """
        raise NotImplementedError("This block does not support backward pass.")

    @property
    def data_g(self):
        if hasattr(self, "_data_g"):
            return self._data_g
        else:
            raise NotImplemented("This block does not have generated data.")


class ProcessingLayer(GenerativeBlock, UpdateBlock):
    """
    An abstract layer for data processing layer in the brain.

    A `ProcessingLayer` is nothing if it does not process data. So every
    sub-class of `ProcessingLayer` should possess a `data` property as
    interface to provide the data it processes.

    For now, all processing layers only have one output, and provide it via
    property `data`, which can be a list to hold multiple data. So it overrides
    the `data` method of `ProcessingBlock`.

    Optionally a `ProcessingLayer` could have a `loss` property for loss (or
    losses) in this layer and a `eval` property for any evaluation metrics in
    this layer. `loss` should be a list if there are multiple losses, so is
    eval graphs.

    If `inputs` in the constructor is not None (in this case it should be a
    list), this layer is supposed to have multiple inputs. Refer to
    `system.GraphSystem` for more explanation.
    """
    _flags = copy.copy(ValidatableProcessingBlock._flags)
    _flags.append("summarize_variables")

    def __init__(self,
                 moving_average_decay=None,
                 inputs=None,
                 summarize_variables=None,
                 **kwargs):
        """
        Args:
            moving_average_decay: A fraction. If `None`, When the parameters of
                this layer is being shared, the shared paras should be
                the current value of the paras. If has a value, it would be
                used in `tf.train.ExponentialMovingAverage`. Then the shared
                value would be the moving average.
            inputs: list
                A list to list inputs of this layer. Refer to
                `system.GraphSystem` for more explanation.
            summarize_variables: bool
                Whether to do tensorboard summary on variables.
        """
        super(ProcessingLayer, self).__init__(**kwargs)

        assert moving_average_decay is None or \
            moving_average_decay >= 0.5 and moving_average_decay < 1, \
            ("Invalid moving_average_decay value {}. Should be None or"
             " between [0.5, 1]".format(moving_average_decay))
        self.moving_average_decay = moving_average_decay

        self.inputs = inputs

        # Bookkeeping all variables.
        self.summarize_variables = summarize_variables
        self.var_list = []

    def set_shadow(self):
        super(ProcessingLayer, self).set_shadow()
        self.var_list = []

    @deprecated(details="This method only worked for tensorflow backends, and is broken now.")
    def _variable_with_weight_decay(self, name, shape, init_para=None):
        """Helper to create an initialized Variable with weight decay.

        Args:
            name: name of the variable
            shape: list of ints

        Returns:
            (Variable Tensor, Weight Decay Loss) If `self.wd` is `None`, then
            the returned weight decay loss would be `None`.
        """
        if not init_para:
            init_para = self.init_para

        var = self._get_variable(name, shape, self._get_initializer(init_para))
        if len(shape) > 1:
            # Add non-bias filters to the collection.
            tf.add_to_collection(FILTER_WEIGHT_COLLECTION, var)

        weight_decay = None
        if self.wd and self.wd["scale"] is not 0:
            try:
                self.log("Using {} regularization with scale {}".format(
                    self.wd["type"], self.wd["scale"]))

                if self.wd["type"] == "l2":
                    weight_decay = A.mul(A.nn.l2_loss(var),
                                         self.wd["scale"],
                                         name=name + '/l2_loss')
                elif self.wd["type"] == "l1":
                    weight_decay = A.mul(A.nn.l1_loss(var),
                                         self.wd["scale"],
                                         name=name + '/l1_loss')
                else:
                    raise Exception("Type {} loss is not supported!".format(
                        self.wd["type"]))
            except KeyError as e:
                raise Exception("`{}` not found in the provided regularization"
                                " parameters, `wd`. Perhaps you have some"
                                " typos.".format(e.message))

        return var, weight_decay

    def _get_initializer(self, init_para=None):
        if not init_para:
            name = "default"
            init = initializers.get(name)
            kwargs = {}
        else:
            name = init_para["name"]
            kwargs = init_para.copy()
            kwargs.pop("name")
            init = initializers.get(name, **kwargs)

        if not self.is_setup:
            self.log("Variables of {} will use initializer {} with arguments {}"
                     " if the variables have not existed yet".format(
                         self.name, name, kwargs))

        return init

    def _pre_setup(self):
        super(ProcessingLayer, self)._pre_setup()

        if self.moving_average_decay:
            # We pass current training step to moving average to speed up
            # updates moving average of variables at the beginning of the
            # training since moving average is useful only later.
            step = A.get_step()
            self.moving_averages = tf.train.ExponentialMovingAverage(
                self.moving_average_decay, step)

    @property
    def outputs(self):
        """
        Return all outputs of this block in a list.

        Outputs that are accessed thorough functions call will not be included.
        """
        outputs = []
        if self.data:
            outputs.extend(self.data) if type(self.data) is list else outputs.append(self.data)
        if self.eval:
            outputs.extend(self.eval)
        if not callable(self.loss):
            if self.loss:
                outputs.extend(self.loss) if type(self.loss) is list else outputs.append(self.loss)

        return outputs

    def _post_setup(self):
        super(ProcessingLayer, self)._post_setup()
        # Maintain moving averages of variables.
        if self.moving_average_decay and len(self.var_list) is not 0:
            self.moving_averages_op = self.moving_averages.apply(self.var_list)

            for var in self.var_list:
                var_average = self.moving_averages.average(var)
                self._var_summary(A.get_name(var) + "_average", var_average)

        if self.do_summary and self.summarize_variables:
            for var in self.var_list:
                self._var_summary(A.get_name(var), var)

        # Log parameter number of this layer.
        total_para_num = 0
        for var in self.var_list:
            shape = A.get_shape(var)
            para_num = 1
            for dim in shape:
                para_num *= dim
            total_para_num += para_num
        self.log("This layer has {} parameters.".format(total_para_num))

    def _post_forward(self, *args, **kwargs):
        # NOTE: this method does not call that of its super class. It has code
        # redundancy, and may be improved.
        for f in self.post_forward_hook:
            f(self, *args, **kwargs)

        # If has done forward for building computational graphs that exist in
        # form of ops, return. But it is possible we still need to build ops
        # for validation block.

        # Do not do summary for shadow copies.
        if self.is_shadow and not self.debug:
            return

        if self.done_first_pass and not self.is_val:
            return

        if not self.do_summary:
            return

        if not self.is_val\
           or (self.is_val\
               and self.do_summary_on_val\
               and not self.done_first_pass_val):
            if self.is_val:
                collections = [VALID_SUMMARY_COLLECTION]
            else:
                collections = [TRAIN_SUMMARY_COLLECTION]

            if self.summarize_output and self.data is not None:
                self._data_summary(self.data, collections=collections)
            if self.loss is not None:
                self._data_summary(self.loss, collections=collections)
            if self.eval is not None:
                self._data_summary(self.eval, collections=collections)


    def _on_update(self, K_prev):
        # Keep propagating the Riemannian metric.
        self.K = K_prev

    def on_para_update(self):
        """
        Operations to run after parameter update.

        Returns:
            A list of ops.
        """
        if hasattr(self, "moving_average_op"):
            return [self.moving_averages_op]
        else:
            return []

    def _var_summary(self, tag, var):
        self.log("Do tensorboard summary on variables {} of {}".format(
            A.get_name(var), self.name))
        if len(A.get_shape(var)) is 0:
            A.summary.scalar(tag,
                             var,
                             collections=[TRAIN_SUMMARY_COLLECTION])
        else:
            A.summary.histogram(tag,
                                var,
                                summary_on_grad=True,
                                collections=[TRAIN_SUMMARY_COLLECTION])

    @property
    def data(self):
        """
        All sub-class `ProcessingLayer` should save the processed data to
        `_data`.
        """
        if hasattr(self, "_data"):
            return self._data
        else:
            return None

    @property
    def loss(self):
        """
        Each `ProcessingLayer` could optionally associated it with a loss. A
        `ProcessingLayer` is the smallest unit in the hierarchical data
        processing system. Since for every data processing system, it would has
        a loss, which is the purpose of its existence, so a `ProcessingLayer`
        layer also could have a loss.

        All sub-classes should save loss to `_loss`.
        """
        if hasattr(self, "_loss"):
            return self._loss
        else:
            return None

    @property
    def eval(self):
        """
        All sub-class `ProcessingLayer` should save the evaluation graph to
        `_eval`.
        """
        if hasattr(self, "_eval"):
            return self._eval
        else:
            return None

    @property
    def verbose_eval(self):
        """
        More verbosely evaluation results. Put the results under
        `_verbose_eval`.
        """
        if hasattr(self, "_verbose_eval"):
            return self._verbose_eval
        else:
            return None

    @property
    def train_op(self):
        """
        Ops that need called along with the top level train op. Sub-class
        should save the train op to `_train_op`.
        """
        if hasattr(self, "_train_op"):
            return self._train_op
        else:
            return None

    def _get_variable(self, name, shape, initializer, trainable=True):
        """
        Allocate or retrieve a trainable variable.

        The following moving average only works in tensorflow backend.  If the
        variable has already existed, depending on `moving_average_decay`'s
        value, moving average of it(when `moving_average_decay` has a value) or
        the original variable(when `moving_average_decay` is None) would be
        returned.
        """
        var = A.get_variable(name,
                             shape,
                             initializer=initializer,
                             trainable=trainable)

        if self.is_setup:
            if self.moving_average_decay:
                log.debug("Use moving average of paras {}".format(
                    A.get_name(var)))
                var = self.moving_averages.average(var)
            else:
                self.log("Reuse paras {}".format(A.get_name(var)))
        else:
            # Append it to the var list, do moving average later in
            # `_post_setup`.
            self.var_list.append(var)

        return var



from __future__ import absolute_import
import tensorflow as tf
import numpy as np

# Data format
# #########################################################################
SUPPORT_PARA_FORMAT = ['oihw', 'hwio']
SUPPORT_DATA_FORMAT = ['nchw', 'nhwc']


# Parallel
# #######################################################################
TIMEOUT = 1



# Backends
# #########################################################################
TF = 'tensorflow'
TORCH = 'pytorch'
_backends = [TF, TORCH]


# Debug flags
# #########################################################################
_debug = False
def set_debug_flag(v):
    global _debug
    _debug = v


def get_debug_flag():
    return _debug


def available_backends():
    return _backends


variable_scope = tf.variable_scope
get_variable_scope = tf.get_variable_scope


def get_scope_name():
    return get_variable_scope().name


# Clock (global step)
# # #######################################################################
# The name assigned to the current training step. It is used to create
# universal access to current training step. This is similar to clock in a
# computer, but in distributed system, such synchronized clock should not
# exist, but it could understand as physical time, meaning how long this
# kid has been trained or born.
_step = 0


def get_step():
    return _step


def step():
    """
    Move to next step
    """
    global _step
    _step += 1


def reset_step():
    global _step
    _step = 0


def set_step(step):
    global _step
    _step = step


_epoch = 0

def get_epoch():
    return _epoch


def inc_epoch():
    global _epoch
    _epoch += 1


def reset_epoch():
    global _epoch
    _epoch = 0

# Computation options
# #########################################################################
# TODO: this option does not work for tensorflow backend yet.
_use_cuda = True


def use_cuda(v=None):
    """
    If `v` is not None, set the value of flag use cuda, otherwise, return the
    flag's value.
    """
    global _use_cuda
    if v is not None:
        _use_cuda = v
    else:
        return _use_cuda

# Naming
# #######################################################################
# A dict to count the number of blocks instantiated. It is used for automatic
# naming.
_block_count = {}


def inc_block_count(block_class):
    try:
        _block_count[block_class] += 1
    except KeyError:
        _block_count[block_class] = 1


def get_block_count(block_class):
    return _block_count[block_class]


def reset_block_count():
    global _block_count
    _block_count = {}


def is_name_the_same(name_1, name_2):
    """
    Return if the name refer to the same named tensor. It seems a trivial
    string match. However, it can check whether two tensors on different devices
    refer to the same tensor, which has different suffixes.
    """
    name_1 = name_1.split(":")[0]
    name_2 = name_2.split(":")[0]
    return name_1 == name_2

# Data type
# #########################################################################
# TODO: data type is not enforced for now. It is possible even the data type is
# set to float32, the actual data being processed is not single precision number.
class DTYPE(object):
    FLOAT32 = 1
    FLOAT64 = 2

dtype = DTYPE.FLOAT32

def set_dtype(d):
    """
    Set data type by passing `DTYPE.FLOAT32`. See class `DTYPE` for supported
    data types.
    """
    global dtype
    dtype = d


def get_np_dtype():
    if dtype == DTYPE.FLOAT32:
        return np.float32
    elif dtype == DTYPE.FLOAT64:
        return np.float64
    else:
        raise ValueError("Date type not supported.")

# Block modes.
# #########################################################################
# Enum types to provide mode a block could be in.
class Mode:
    TRAIN = "train"
    VAL = "val"
    TEST = "test"


def check_mode(mode):
    if mode != Mode.TRAIN \
        and mode != Mode.VAL \
        and mode != Mode.TEST:
        raise ValueError("Mode `{}` is not support.".format(mode))

# Named objects.
# #########################################################################


class NamedValue(object):
    pass


class NamedTensorTuple(tuple, NamedValue):
    """
    Object to pass a tuple of tensors around. It enables the tuple to be named,
    thus, be identifiable by name.

    To create a `NamedTensorTuple`::

        t1, t2 = ... # Two tensors
        NamedTensorTuple("any_name", t1, t2)

    Arbitrary number of tuple members are supported.
    """
    def __init__(self, name=None, *args, **kwargs):
        assert type(name) is str, "Named should be string."
        self.name = name

    def __new__(cls, name, *args, **kwargs):
        return super(NamedTensorTuple, cls).__new__(cls, *args, **kwargs)


class NamedNumericValue(NamedValue):
    pass


class NamedScalar(np.float, NamedNumericValue):
    """
    Object to pass a scalar. It enables the scalar to be named, thus, be
    identifiable by name.

    To create a `NamedScalar`::

        s = ... # A scalar
        NamedScalar("any_name", s)
    """
    def __init__(self, name=None, *args, **kwargs):
        self.name = name

    def __new__(cls, name, *args, **kwargs):
        return super(NamedScalar, cls).__new__(cls, *args, **kwargs)


class NamedNdarray(np.ndarray, NamedNumericValue):
    """
    Object to pass a named numpy array.

    To create a `NamedNdarray`::

        a = ... # A numpy array
        NamedNdarray("any_name", a)

    Refer to https://docs.scipy.org/doc/numpy/user/basics.subclassing.html for
    how to subclass numpy array.
    """
    def __new__(cls, name, array):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        # It also triggers a call to NamedNdarray.__array_finalize__
        obj = np.asarray(array).view(cls)
        # set the new 'info' attribute to the value passed
        obj.name = name
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # ``self`` is a new object resulting from
        # ndarray.__new__(NamedNdarray, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. NamedNdarray():
        #    obj is None
        #    (we're in the middle of the InfoArray.__new__
        #    constructor, and self.info will be set when we return to
        #    InfoArray.__new__)
        if obj is None: return
        # From view casting - e.g arr.view(NamedNdarray):
        #    obj is arr
        #    (type(obj) can be NamedNdarray)
        # From new-from-template - e.g namedarr[:3]
        #    type(obj) is NamedNdarray
        #
        # Note that it is here, rather than in the __new__ method,
        # that we set the default value for 'info', because this
        # method sees all creation of default objects - with the
        # NamedNdarray.__new__ constructor, but also with
        # arr.view(NamedNdarray).
        self.name = getattr(obj, 'name', None)
        # We do not need to return anything



from __future__ import absolute_import
from .. import backend as A
from ..utils.tools import is_tuple_or_list
import torch as th
import tensorflow as tf

import threading
from six.moves import map
from six.moves import range
from six.moves import zip


_lock = threading.Lock()


def scatter(data, devices):
    """
    Split the data of a list of data at the first dimension by devices number,
    and send data respectively to devices.
    """
    def scatter_inner(data):
        if type(data) is list:
            return tuple(zip(*list(map(scatter_inner, data))))
        else:
            return A.scatter_parallel(data, devices)

    with A.variable_scope("data_scatter"):
        return scatter_inner(data)


def gather(data, output_device=0):
    def gather_inner(data):
        d = data[0]
        if is_tuple_or_list(d):
            return list(map(gather_inner, list(zip(*data))))
        else:
            return A.gather_parallel(data, output_device)

    return gather_inner(data)


def device(device_id):
    if A.backend() == A.TORCH:
        # remove the device type prefix and return
        return th.cuda.device(int(device_id[-1]))
    elif A.backend() == A.TF:
        return tf.device(device_id)
    else:
        raise ValueError("Backend is not supported.")


def broadcast(data, devices):
    """
    Broadcast variables in `data` to `devices`. The results are kept in the
    'globally', which can be retried using `get_variable`.
    """
    if A.torch_version < 0.4:
        out = th.nn.parallel._functions.Broadcast(devices)(*data)
    else:
        out = th.nn.parallel._functions.Broadcast.apply(devices, *data)
    # Deflatten the data.
    if len(data) > 0:
        out = [out[i:i + len(data)] for i in range(0, len(out), len(data))]

    if A.get_name(data[0]) is not None:
        device_num = len(devices)
        for i in range(1, device_num):
            with device("/gpu:{}".format(i)):
                for j, t in enumerate(out[i]):
                    name = "{}:{}".format(A.get_name(data[j]).split(':')[0], i)
                    A.cache_tensor(t, name)

    return out


def thread_run(func, *args, **kwargs):
    """
    Run code in a separate thread. Arbitrary positional arguments and keywords
    arguments are supported.

    Args:
        func: the function object to run

    Returns:
        results: list
            A list that contains returned results of the `func`. Empty if none.
        thread: threading.Thread
            The thread that is created for running.
    """
    # Give that new thread loses context manager, enforce it again.
    current_device = str(th.cuda.current_device())
    def _worker(func, results, *args, **kwargs):
        with device("/gpu:{}".format(current_device)):
            ret = func(*args, **kwargs)
            with _lock:
                results.append(ret)

    results = []
    args = [a for a in args]
    args.insert(0, results)
    args.insert(0, func)
    thread = threading.Thread(target=_worker,
                              args=tuple(args),
                              kwargs=kwargs)
    thread.start()
    return results, thread



from __future__ import division


from __future__ import absolute_import
import torch as th
from torch.nn import functional as F
from torch import nn as tnn

from .computational_graph import cache_name_if_exist
from . import computational_graph as cg
from akid.utils import glog as log


@cache_name_if_exist
def l2_loss(var, name=None):
    return th.div(th.sum(var * var), 2)


@cache_name_if_exist
def l2_norm(var, dim=None, name=None):
    """
    Args:
         var: tensor
             tensor of rank 1 or 2. If the rank is 1, it is taken as a vector,
             and norm computed. If 2, it is taken as an array of vectors, where
             var[0] is a vector. The norm of each individually is computed.
         dim: the dimension along which to compute l2 norm. Not useful when
             `var` is of one dimension.
    """
    shape = cg.get_shape(var)
    if len(shape) == 1 or dim is None:
        return th.sqrt(th.sum(var * var))
    else:
        return th.sqrt(th.sum(var * var, dim=dim))


@cache_name_if_exist
def l1_loss(var, name=None):
    return th.sum(th.abs(var))


@cache_name_if_exist
def conv2d(input, filter, bias=None, strides=1, padding=0, name=None):
    shape = cg.get_shape(filter)
    H, W = shape[-2], shape[-1]
    shape = cg.get_shape(input)
    H_in, W_in = shape[-2], shape[-1]
    if type(padding) is str:
        padding = padding_str2tuple(H_in, W_in, strides, padding, H, W)
    strides = _normalize_stride(strides)

    if len(padding) == 4:
        # We need to do manual padding since by default, conv2d only do
        # two-side padding with the same padding dim.
        padding = list(padding)
        padding.reverse()
        padding = tuple(padding)
        input = th.nn.functional.pad(input, padding)
        padding = 0

    return F.conv2d(input, filter, bias, stride=strides, padding=padding)


@cache_name_if_exist
def conv1d(x, W, b=None, stride=1, padding=0, name=None):
    return F.conv1d(x, W, b, stride, padding)


@cache_name_if_exist
def inner_product(input, W, bias=None, name=None):
    if bias is not None:
        return th.addmm(bias, input, W)
    else:
        return input.matmul(W)


@cache_name_if_exist
def bmm(M1, M2, name=None):
        return th.bmm(M1, M2)


@cache_name_if_exist
def max_pool(value, ksize, strides, padding, data_format="NHWC", return_indices=False, name=None):
    ksize = _normalize_ksize(ksize)
    H, W = ksize[0], ksize[1]
    shape = cg.get_shape(value)
    H_in, W_in = shape[-2], shape[-1]
    padding = padding_str2tuple(H_in, W_in, strides, padding, H, W)
    # Only take the padding at the beginning, since PyTorch by default only pad
    # the same at the both end.
    if type(padding) is not int and len(padding) == 4:
        padding = (padding[0], padding[2])
    strides = _normalize_stride(strides)
    return F.max_pool2d(value, ksize, strides, padding, return_indices=return_indices)


@cache_name_if_exist
def max_pool1d(x, ksize, stride, padding, name=None):
    return F.max_pool1d(x, ksize, stride, padding)


def _normalize_stride(strides):
    # The format of ksize in torch is a tuple of size 2 instead of 4 in
    # tensorflow.
    if type(strides) is int:
        strides = (strides, strides)
    elif len(strides) == 4:
        # log.warning("Torch backend does not support stride in all four dimensions."
        #             " Use the middle two as height and width.")
        strides=(strides[-3], strides[-2])
    elif type(strides) is not tuple:
        strides = tuple(strides)

    return strides

_normalize_ksize = _normalize_stride


@cache_name_if_exist
def avg_pool(value, ksize, strides, padding, name=None):
    ksize = _normalize_ksize(ksize)
    H, W = ksize[0], ksize[1]
    shape = cg.get_shape(value)
    H_in, W_in = shape[-2], shape[-1]
    padding = padding_str2tuple(H_in, W_in, strides, padding, H, W)
    strides = _normalize_stride(strides)
    return F.avg_pool2d(value, ksize, strides, padding)


@cache_name_if_exist
def relu(v, name=None):
    return F.relu(v)


@cache_name_if_exist
def sigmoid(v, name=None):
    return th.sigmoid(v)


@cache_name_if_exist
def embedding(idxs, weights, name=None):
    return F.embedding(idxs, weights)


@cache_name_if_exist
def dropout(v, keep_prob, val=False, in_place=False, name=None):
    return F.dropout(v, 1-keep_prob, training=not val, inplace=in_place)


@cache_name_if_exist
def zero_fraction(data, name=None):
    return th.mean((data == 0).float())


@cache_name_if_exist
def mse_loss(data, labels, size_average=True, name=None):
    return th.nn.functional.mse_loss(data, labels, reduction="mean")


@cache_name_if_exist
def cross_entropy_loss(logits, labels, name=None):
    return F.cross_entropy(logits, labels)


@cache_name_if_exist
def margin_ranking_loss(x1, x2, target, margin=0, name=None):
    return F.margin_ranking_loss(x1, x2, target, margin=margin)


@cache_name_if_exist
def consine_similarity(x1, x2, dim=1, eps=1e-8):
    return F.cosine_similarity(x1, x2, dim, eps)


@cache_name_if_exist
def binary_cross_entropy_loss_with_logits(logits, labels, pos_weight=None, name=None):
    return F.binary_cross_entropy_with_logits(logits, labels, pos_weight=pos_weight)


@cache_name_if_exist
def class_acccuracy(predictions, labels, name=None):
    size = cg.get_shape(labels)[0]
    pred = predictions.max(1, keepdim=True)[1] # get the index of the max log-probability
    correct = pred.eq(labels.view_as(pred)).sum()
    acc = correct.float() / size
    return acc


@cache_name_if_exist
def normalize_weight(W):
    """
    Normalize the norm of each weight that corresponds to an output channel to 1.
    """
    shape = cg.get_shape(W)
    if len(shape) == 4:
        c_out, c_in, h, w = shape
        W = W.view(c_out, c_in*h*w)

    norm = l2_norm(W)
    W_normalized = W / norm[:, None]

    if len(shape) == 4:
        W_normalized = W_normalized.view(shape)

    return W_normalized


@cache_name_if_exist
def nn_riemannic_metric(K, W, b, need_to_normalize_weights=False):
    """
    Given Riemannian metric of the previous layer, compute that of this layer.

    Due to the fact that computing similarity between group action generated
    filters would require backtracking computation of similarity of all spatial
    location in the image, which is not feasible, only similarity between
    filters of different channels are used.

    Args:
        K: the Riemannian metric
        W: the weigth matrix, 4D or 2D, or None (means identity matrix)
        b: the bias vector
    """
    shape = cg.get_shape(W)

    if len(shape) == 4:
        c_out, c_in, h, w = shape
        W = W.permute(2, 3, 0, 1).contiguous()
        W = W.view(h*w, c_out, c_in)
        W_T = th.transpose(W, 1, 2)
        # Compute sum_{H * W}(WKW^{T}_{i})
        if K is not None:
            left = th.matmul(W, K)
        else:
            left = W
        if b is not None:
            mat = th.ger(b, b)  # Outer product
            K_out = th.addbmm(mat, left, W_T)
        else:
            K_out = th.sum(th.bmm(left, W_T), dim=0)

    elif len(shape) == 2:
        if K is not None:
            right = th.matmul(K, W)
        else:
            right = W

        if b is not None:
            mat = th.ger(b, b)  # Outer product
            K_out = th.addmm(mat, W.t(), right)
        else:
            K_out = th.mm(W.t(), right)
    else:
        raise ValueError("Shape {} is not supported".format(shape))

    K_out = K_out * (K_out > 0).float()

    return K_out


def padding_str2tuple(H_in, W_in, strides, padding, H, W):
    """
    Convert padding from string to tuple. Thea meaning of string is from
    tensorflow.
    """
    if type(strides) is int:
        strides = (strides, strides)

    if padding == 'VALID':
        padding = 0
    elif padding == 'SAME':
        H_pad = get_padding_SAME(H_in, strides[0], H)
        W_pad = get_padding_SAME(W_in, strides[1], W)
        if H_pad[0] == H_pad[1] and W_pad[0] == W_pad[1]:
            # Use plan to use the default padding in conv op.
            padding = (H_pad[0], W_pad[0])
        else:
            padding = (*H_pad, *W_pad)
    else:
        raise ValueError("{} padding is not supported. Should be VALID or SAME".format(padding))

    return padding


def get_padding_SAME(input_size, stride, ksize):
    out_size = (input_size + stride - 1) // stride
    padding_needed = max(0, (out_size - 1) * stride + ksize - input_size)
    padding_before = padding_needed // 2
    padding_after = padding_needed - padding_before
    # Since torch seems not to support 4 tuple padding, just return the left part
    padding = (padding_before, padding_after)

    return padding

@cache_name_if_exist
def bn(x, out_channel_num, # out_channel_num is unused.
       gamma, beta,
       running_mean=None, running_var=None,
       is_val=False, track_running_stats=True,
       exponential_average_factor=0.1, eps=1^-5,
       name=None, **kwargs):
    return F.batch_norm(
        x, running_mean, running_var, gamma, beta,
        not is_val or not track_running_stats,
        exponential_average_factor, eps)


@cache_name_if_exist
def hessian(l, x, name=None):
    """
    Compute the Hessian of l w.r.t. x, where l is the output of a loss function.
    """
    nabla = th.autograd.grad(l, x, create_graph=True)
    nabla = th.cat([g.contiguous().view(-1) for g in nabla])
    H = []
    for p in nabla:
        p = th.cat(th.autograd.grad(p, x, retain_graph=True))
        H.append(p)

    H = th.stack(H, 1)
    return H


@cache_name_if_exist
def grad(l, x, flatten=False, name=None):
    """
    If `flatten` is True, the output grad would be flattened to a vector.
    """
    nabla = th.autograd.grad(l, x, create_graph=True)

    if flatten:
        if type(nabla) is tuple:
            nabla = th.cat([g.contiguous().view(-1) for g in nabla])
        elif type(nabla) is th.Tensor:
            nabla = nabla.contiguous().view(-1)
        else:
            raise ValueError("Type {} of grad is not supported".format(type(nabla)))

    return nabla

@cache_name_if_exist
def hessian_vector_product(nabla, x, v, allow_unused=False, name=None):
    """
    Take a gradient `nabla` computed by PyTorch and compute the Hessian vector
    product between it and a vector `v`. To compute Hessian from `grad`, we
    need the parameters `x` that are supposed to be taken derivatives against.
    """
    Hv = th.autograd.grad(nabla, x, allow_unused=allow_unused, grad_outputs=v)
    # import ipdb; ipdb.set_trace()
    Hv = th.cat([p.contiguous().view(-1) for p in Hv])
    return Hv



"""
Source for ImageNet data set.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from abc import ABCMeta
from abc import abstractmethod
import os

import tensorflow as tf
from torchvision import datasets, transforms

from akid import (
    SupervisedSource,
    ClassificationTFSource,
    StaticSource
)

from akid import backend as A
import six


"""Small library that points to a data set.

Methods of Data class:
  data_files: Returns a python list of all (sharded) data set files.
  num_examples_per_epoch: Returns the number of examples in the data set.
  num_classes: Returns the number of classes in the data set.
  reader: Return a reader for a single entry from the data set.
"""


class Dataset(six.with_metaclass(ABCMeta, object)):
    """A simple class for handling data sets."""

    def __init__(self, work_dir, name, subset):
        """Initialize dataset using a subset and the path to the data."""
        assert subset in self.available_subsets(), self.available_subsets()
        self.work_dir = work_dir
        self.name = name
        self.subset = subset

    @abstractmethod
    def num_classes(self):
        """Returns the number of classes in the data set."""
        pass
        # return 10

    @abstractmethod
    def num_examples_per_epoch(self):
        """Returns the number of examples in the data subset."""
        pass
        # if self.subset == 'train':
        #   return 10000
        # if self.subset == 'validation':
        #   return 1000

    @abstractmethod
    def download_message(self):
        """Prints a download message for the Dataset."""
        pass

    def available_subsets(self):
        """Returns the list of available subsets."""
        return ['train', 'validation']

    def data_files(self):
        """Returns a python list of all (sharded) data subset files.

        Returns:
            python list of all (sharded) data set files.
        Raises:
            ValueError: if there are not data_files matching the subset.
        """
        tf_record_pattern = os.path.join(self.work_dir, '%s-*' % self.subset)
        data_files = tf.gfile.Glob(tf_record_pattern)
        if not data_files:
            print('No files found for dataset %s/%s at %s' % (self.name,
                                                              self.subset,
                                                              self.work_dir))

            self.download_message()
            exit(-1)

        return data_files

    def reader(self):
        """Return a reader for a single entry from the data set.

        See io_ops.py for details of Reader class.

        Returns:
        Reader object that reads the data set.
        """
        return tf.TFRecordReader()


class ImagenetData(Dataset):
    """ImageNet data set."""

    def num_classes(self):
        """Returns the number of classes in the data set."""
        return 1000

    def num_examples_per_epoch(self):
        """Returns the number of examples in the data set."""
        # Bounding box data consists of 615299 bounding boxes for 544546
        # images.
        if self.subset == 'train':
            return 1281167
            if self.subset == 'validation':
                return 50000

    def download_message(self):
        """Instruction to download and extract the tarball from Flowers
        website."""

        print('Failed to find any ImageNet %s files' % self.subset)
        print('')
        print('If you have already downloaded and processed the data, then make '
              'sure to set --data_dir to point to the directory containing the '
              'location of the sharded TFRecords.\n')
        print('If you have not downloaded and prepared the ImageNet data in the '
              'TFRecord format, you will need to do this at least once. This '
              'process could take several hours depending on the speed of your '
              'computer and network connection\n')
        print('Please see README.md for instructions on how to build '
              'the ImageNet dataset using download_and_preprocess_imagenet.\n')
        print('Note that the raw data size is 300 GB and the processed data size '
              'is 150 GB. Please ensure you have at least 500GB disk space.')


class ImagenetTFSource(ClassificationTFSource):
    def __init__(self, has_super_label=True, **kwargs):
        super(ImagenetTFSource, self).__init__(**kwargs)
        self.has_super_label = has_super_label

    def _forward(self):
        """
        Read, and crop images using bounding boxes. The cropped images and its
        labels are the output of this source. In this case of validation
        samples, no cropping is done.
        """

        with tf.device('/cpu:0'):
            dataset = ImagenetData(work_dir=self.work_dir,
                                   subset="train",
                                   name="Imagenet")
            tensor_list = self._read_dataset(dataset, train=True)
            self._training_datum = tensor_list[0]
            self._training_label = tensor_list[1:]
            # Manually set shape value, which is necessary for later
            # usage. The height and width info is dynamically determined when
            # decoding jpg files, so is not available. tf.slice seems to lose
            # all shape info when one of the shape info is missing before
            # slicing.
            self._training_datum.set_shape([None, None, 3])
            for t in self._training_label:
                t.set_shape([1])

            dataset = ImagenetData(work_dir=self.work_dir,
                                   subset="validation",
                                   name="Imagenet")
            tensor_list = self._read_dataset(dataset, train=False)
            self._val_datum = tensor_list[0]
            self._val_label = tensor_list[1:]
            self._val_datum.set_shape([None, None, 3])
            for t in self._val_label:
                t.set_shape([1])

    def _read_dataset(self, dataset, train):
        """
        Given a dataset return the data tensors for images and labels.
        """
        data_files = dataset.data_files()
        if data_files is None:
            raise ValueError('No data files found for this dataset')

        # Create filename_queue
        if train:
            filename_queue = tf.train.string_input_producer(data_files,
                                                            shuffle=True,
                                                            capacity=16)
        else:
            filename_queue = tf.train.string_input_producer(data_files,
                                                            shuffle=False,
                                                            capacity=1)
        reader = dataset.reader()
        _, example_serialized = reader.read(filename_queue)
        image_buffer, label, bbox, _ = self.parse_example_proto(
            example_serialized)
        if train:
            image = self.process_train_image(image_buffer, bbox, 0)
        else:
            image = self.decode_jpeg(image_buffer)
        if type(label) is list:
            label.insert(0, image)
            out_tensor_list = label
        else:
            out_tensor_list = [image, label]

        return out_tensor_list

    def decode_jpeg(self, image_buffer, scope=None):
        """Decode a JPEG string into one 3-D float image Tensor.

        Args:
            image_buffer: scalar string Tensor.
            scope: Optional scope for op_scope.
        Returns:
            3-D float Tensor with values ranging from [0, 1).
        """
        with tf.name_scope(values=[image_buffer], name=scope, default_name='decode_jpeg'):
            # Decode the string as an RGB JPEG.
            # Note that the resulting image contains an unknown height and
            # width that is set dynamically by decode_jpeg. In other words, the
            # height and width of image is unknown at compile-time.
            image = tf.image.decode_jpeg(image_buffer, channels=3)

            # After this point, all image pixels reside in [0,1) until the very
            # end, when they're rescaled to (-1, 1).  The various adjust_* ops
            # all require this range for dtype float.
            image = tf.image.convert_image_dtype(image, dtype=tf.float32)
            return image

    def parse_example_proto(self, example_serialized):
        """Parses an Example proto containing a training example of an image.

        The output of the build_image_data.py image preprocessing script is a
        dataset containing serialized Example protocol buffers. Each Example
        proto contains the following fields:

        image/height: 462
        image/width: 581
        image/colorspace: 'RGB'
        image/channels: 3
        image/class/label: 615
        image/class/synset: 'n03623198'
        image/class/text: 'knee pad'
        image/object/bbox/xmin: 0.1
        image/object/bbox/xmax: 0.9
        image/object/bbox/ymin: 0.2
        image/object/bbox/ymax: 0.6
        image/object/bbox/label: 615
        image/format: 'JPEG'
        image/filename: 'ILSVRC2012_val_00041207.JPEG'
        image/encoded: <JPEG encoded string>

        Args:
            example_serialized: scalar Tensor tf.string containing a serialized
                Example protocol buffer.

        Returns:
            image_buffer: Tensor tf.string containing the contents of a JPEG
                file.
            label: Tensor tf.int32 containing the label.
            bbox: 3-D float Tensor of bounding boxes arranged [1, num_boxes,
                coords] where each coordinate is [0, 1) and the coordinates are
                arranged as [ymin, xmin, ymax, xmax].
            text: Tensor tf.string containing the human-readable label.
        """
        # Dense features in Example proto.
        feature_map = {
            'image/encoded': tf.FixedLenFeature([], dtype=tf.string,
                                                default_value=''),
            'image/class/label': tf.FixedLenFeature([1], dtype=tf.int64,
                                                    default_value=-1),
            'image/class/text': tf.FixedLenFeature([], dtype=tf.string,
                                                   default_value=''),
        }
        if self.has_super_label:
            feature_map['image/class/super_label'] \
                = tf.FixedLenFeature([1], dtype=tf.int64, default_value=-1)
        sparse_float32 = tf.VarLenFeature(dtype=tf.float32)
        # Sparse features in Example proto.
        feature_map.update(
            {k: sparse_float32 for k in ['image/object/bbox/xmin',
                                         'image/object/bbox/ymin',
                                         'image/object/bbox/xmax',
                                         'image/object/bbox/ymax']})

        features = tf.parse_single_example(example_serialized, feature_map)
        label = tf.cast(features['image/class/label'], dtype=tf.int32)
        if self.has_super_label:
            super_label = tf.cast(features['image/class/super_label'],
                                dtype=tf.int32)

        xmin = tf.expand_dims(features['image/object/bbox/xmin'].values, 0)
        ymin = tf.expand_dims(features['image/object/bbox/ymin'].values, 0)
        xmax = tf.expand_dims(features['image/object/bbox/xmax'].values, 0)
        ymax = tf.expand_dims(features['image/object/bbox/ymax'].values, 0)

        # Note that we impose an ordering of (y, x) just to make life
        # difficult.
        bbox = tf.concat(axis=0, values=[ymin, xmin, ymax, xmax])

        # Force the variable number of bounding boxes into the shape
        # [1, num_boxes, coords].
        bbox = tf.expand_dims(bbox, 0)
        bbox = tf.transpose(bbox, [0, 2, 1])

        if self.has_super_label:
            return features['image/encoded'],\
                [label, super_label],\
                bbox,\
                features['image/class/text']
        else:
            return features['image/encoded'],\
                label,\
                bbox,\
                features['image/class/text']

    def process_train_image(self, image_buffer, bbox, thread_id):
        # Crop data
        image = self.decode_jpeg(image_buffer)
        # Each bounding box has shape [1, num_boxes, box coords] and
        # the coordinates are ordered [ymin, xmin, ymax, xmax].

        # If preparing for training set, we crop the image by the
        # bounding boxes (if available).
        # if train:
        # Display the bounding box in the first thread only.
        if not thread_id:
            image_with_box = tf.image.draw_bounding_boxes(
                tf.expand_dims(image, 0), bbox)
            tf.summary.image('image_with_bounding_boxes', image_with_box)

        # A large fraction of image datasets contain a human-annotated
        # bounding box delineating the region of the image containing
        # the object of interest.  We choose to create a new bounding
        # box for the object which is a randomly distorted version of
        # the human-annotated bounding box that obeys an allowed range
        # of aspect ratios, sizes and overlap with the human-annotated
        # bounding box. If no box is supplied, then we assume the
        # bounding box is the entire image.
        sample_distorted_bounding_box \
            = tf.image.sample_distorted_bounding_box(
                tf.shape(image),
                bounding_boxes=bbox,
                min_object_covered=0.1,
                aspect_ratio_range=[0.75, 1.33],
                area_range=[0.05, 1.0],
                max_attempts=100,
                use_image_if_no_bounding_boxes=True)
        bbox_begin, bbox_size, distort_bbox \
            = sample_distorted_bounding_box
        if not thread_id:
            image_with_distorted_box = tf.image.draw_bounding_boxes(
                tf.expand_dims(image, 0), distort_bbox)
            tf.summary.image('images_with_distorted_bounding_box',
                             image_with_distorted_box)

        # Crop the image to the specified bounding box.
        image = tf.slice(image, bbox_begin, bbox_size)

        return image


class ImagenetTorchSource(StaticSource, SupervisedSource):
    def __init__(self, random_sized_crop=True, **kwargs):
        super(ImagenetTorchSource, self).__init__(**kwargs)
        self.random_sized_crop = random_sized_crop

    def _setup(self):
        mean=[0.485, 0.456, 0.406]
        std=[0.229, 0.224, 0.225]
        normalize = transforms.Normalize(mean=mean, std=std)
        A.summary.set_normalization(mean, std)

        t_list = []
        if self.random_sized_crop:
            t_list.append(transforms.RandomSizedCrop(224))
        else:
            t_list.append(transforms.Scale(256))
            t_list.append(transforms.RandomCrop(224))

        t_list.append(transforms.RandomHorizontalFlip())
        t_list.append(transforms.ToTensor())
        t_list.append(normalize)

        self.dataset = datasets.ImageFolder(
            self.work_dir + "/train",
            transforms.Compose(t_list))
        self.val_dataset = datasets.ImageFolder(
            self.work_dir + "/val",
            transforms.Compose([
                transforms.Scale(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                normalize,
            ]))
    def _forward(self):
        pass



"""
This module contains utilities to transform tensorflow style objects to Caffe
style.
"""

from __future__ import absolute_import
import sys
import numpy as np
from . import glog as log


def tf_to_caffe_blob(blob):
    """
    Transform tensorflow style numpy array to Caffe blob style.

    Args:
        blob: numpy.array
            tensorflow style numpy array.

    Return:
        Transformed numpy array.
    """
    if len(blob.shape) is 4:
        blob = np.einsum('hwcn->nchw', blob)
    elif len(blob.shape) is 2:
        blob = np.einsum('nc->cn', blob)
    elif len(blob.shape) is 3:
        blob = np.einsum('hwc->chw', blob)
    else:
        log.error("{} shape numpy array is not supported!".format(blob.shape))
        sys.exit()
    return blob



from __future__ import absolute_import
from __future__ import print_function
import os
import six.moves.urllib.request, six.moves.urllib.parse, six.moves.urllib.error
import gzip
import zipfile

import numpy as np

from torchvision import datasets, transforms

from ..core.sources import Source, InMemoryFeedSource, SupervisedSource, StaticSource
from .datasets import DataSet, DataSets
from .. import backend as A


class MNISTFeedSource(InMemoryFeedSource, SupervisedSource):
    """
    A concrete `Source` for MNIST dataset.
    """
    @property
    def shape(self):
        return [28, 28, 1]

    @property
    def label_shape(self):
        return [1]

    def _maybe_download(self, filename, work_directory):
        """Download the data from Yann's website, unless it's already here."""
        if not os.path.exists(work_directory):
            os.mkdir(work_directory)
        filepath = os.path.join(work_directory, filename)
        if not os.path.exists(filepath):
            filepath, _ = six.moves.urllib.request.urlretrieve(self.url + filename, filepath)
            statinfo = os.stat(filepath)
            self.log('Successfully downloaded {} {} bytes.'.format(filename, statinfo.st_size))
        return filepath

    def _load(self, fake_data=False, one_hot=False):
        if fake_data:
            training_dataset = DataSet([], [], fake_data=True)
            validation_dataset = DataSet([], [], fake_data=True)
            test_dataset = DataSet([], [], fake_data=True)
            return DataSets(training_dataset, validation_dataset, test_dataset)

        TRAIN_IMAGES = 'train-images-idx3-ubyte.gz'
        TRAIN_LABELS = 'train-labels-idx1-ubyte.gz'
        TEST_IMAGES = 't10k-images-idx3-ubyte.gz'
        TEST_LABELS = 't10k-labels-idx1-ubyte.gz'

        local_file = self._maybe_download(TRAIN_IMAGES, self.work_dir)
        train_images = self._extract_images(local_file)

        local_file = self._maybe_download(TRAIN_LABELS, self.work_dir)
        train_labels = self._extract_labels(local_file, one_hot=one_hot)

        local_file = self._maybe_download(TEST_IMAGES, self.work_dir)
        test_images = self._extract_images(local_file)

        local_file = self._maybe_download(TEST_LABELS, self.work_dir)
        test_labels = self._extract_labels(local_file, one_hot=one_hot)

        VALIDATION_SIZE = self.validation_rate * self.num_train
        validation_images = train_images[:VALIDATION_SIZE]
        validation_labels = train_labels[:VALIDATION_SIZE]
        train_images = train_images[VALIDATION_SIZE:]
        train_labels = train_labels[VALIDATION_SIZE:]

        training_dataset = DataSet(train_images,
                                   train_labels,
                                   center=self.center,
                                   scale=self.scale)
        validation_dataset = DataSet(validation_images,
                                     validation_labels,
                                     center=self.center,
                                     scale=self.scale)
        test_dataset = DataSet(test_images,
                               test_labels,
                               shuffle=False,
                               center=self.center,
                               scale=self.scale)

        return DataSets(training_dataset, test_dataset, validation_dataset)

    def _read32(self, bytestream):
        dt = np.dtype(np.uint32).newbyteorder('>')
        return int(np.frombuffer(bytestream.read(4), dtype=dt))

    def _dense_to_one_hot(self, labels_dense, num_classes=10):
        """Convert class labels from scalars to one-hot vectors."""
        num_labels = labels_dense.shape[0]
        index_offset = np.arange(num_labels) * num_classes
        labels_one_hot = np.zeros((num_labels, num_classes))
        labels_one_hot.flat[index_offset + labels_dense.ravel()] = 1
        return labels_one_hot

    def _extract_images(self, filename):
        """
        Extract the images into a 4D uint8 np array [index, y, x, depth].
        """
        self.log('Extracting', filename)
        with gzip.open(filename) as bytestream:
            magic = self._read32(bytestream)
            if magic != 2051:
                raise ValueError(
                    'Invalid magic number %d in MNIST image file: %s' %
                    (magic, filename))
            num_images = self._read32(bytestream)
            rows = self._read32(bytestream)
            cols = self._read32(bytestream)
            buf = bytestream.read(rows * cols * num_images)
            data = np.frombuffer(buf, dtype=np.uint8)
            data = data.reshape(num_images, rows, cols, 1)
            return data

    def _extract_labels(self, filename, one_hot=False):
        """Extract the labels into a 1D uint8 numpy array [index]."""
        print(('Extracting', filename))
        with gzip.open(filename) as bytestream:
            magic = self._read32(bytestream)
            if magic != 2049:
                raise ValueError(
                    'Invalid magic number %d in MNIST label file: %s' %
                    (magic, filename))
            num_items = self._read32(bytestream)
            buf = bytestream.read(num_items)
            labels = np.frombuffer(buf, dtype=np.uint8)
            if one_hot:
                return self._dense_to_one_hot(labels)
            return labels


class RotatedMNISTFeedSource(InMemoryFeedSource, SupervisedSource):
    """
    A concrete `Source` for rotated MNIST dataset.
    """
    @property
    def shape(self):
        return [28, 28, 1]

    @property
    def label_shape(self):
        return [1]

    def _maybe_download_and_extract(self, filename):
        """
        Download the data from Yann's website, unless it's already here.

        File contents will be extracted in the `self.work_dir`.
        """
        if not os.path.exists(self.work_dir):
            os.mkdir(self.work_dir)
        filepath = os.path.join(self.work_dir, filename)
        if not os.path.exists(filepath):
            filepath, _ = six.moves.urllib.request.urlretrieve(self.url + filename, filepath)
            statinfo = os.stat(filepath)
            self.log('Successfully downloaded', filename, statinfo.st_size,
                     'bytes.')
            self.log('Extracting zip file ... ')
            f = zipfile.ZipFile(filepath)
            f.extractall(path=self.work_dir)
            self.log('Extraction finished ... ')

    def _load(self, fake_data=False, one_hot=False):
        filename = self.url.split('/')[-1]
        filepath = os.path.join(self.work_dir, filename)
        if not os.path.exists(filepath):
            self._maybe_download_and_extract(filename)

        TRAIN_DATA_PATH = os.path.join(
            self.work_dir,
            'mnist_all_rotation_normalized_float_train_valid.amat')
        TEST_DATA_PATH = os.path.join(
            self.work_dir,
            'mnist_all_rotation_normalized_float_test.amat')

        train_data = np.loadtxt(TRAIN_DATA_PATH)
        test_data = np.loadtxt(TEST_DATA_PATH)
        training_images = np.reshape(train_data[:self.num_train, 0:-1],
                                     [-1, 28, 28, 1])
        training_labels = train_data[:self.num_train, -1]
        test_images = np.reshape(test_data[:self.num_val, 0:-1],
                                 [-1, 28, 28, 1])
        test_labels = test_data[:self.num_val, -1]

        training_dataset = DataSet(training_images,
                                   training_labels,
                                   center=self.center,
                                   scale=self.scale)
        test_dataset = DataSet(test_images,
                               test_labels,
                               center=self.center,
                               scale=self.scale)

        return DataSets(training_dataset, test_dataset)


class MNISTTorchSource(StaticSource, SupervisedSource):
    def _setup(self):
        self.dataset = datasets.MNIST(self.work_dir, train=True, download=True,
                                      transform=transforms.Compose([
                                          transforms.ToTensor(),
                                          transforms.Normalize((0.1307,), (0.3081,))]))
        self.val_dataset = datasets.MNIST(self.work_dir, train=False,
                                      transform=transforms.Compose([
                                          transforms.ToTensor(),
                                          transforms.Normalize((0.1307,), (0.3081,))]))
    def _forward(self):
        pass


class MNISTSource(Source):
    def _setup(self):
        if not hasattr(self, "_data"):
            self._data = datasets.MNIST(self.work_dir, train=True, download=True,
                                        transform=transforms.Compose([
                                            transforms.ToTensor(),
                                            transforms.Normalize((0.1307,), (0.3081,))]))

    def set_mode(self, mode):
        super(MNISTSource, self).set_mode(mode)
        if mode == A.Mode.TRAIN:
            self._data = datasets.MNIST(self.work_dir, train=True, download=True,
                                        transform=transforms.Compose([
                                            transforms.ToTensor(),
                                            transforms.Normalize((0.1307,), (0.3081,))]))
        elif mode == A.Mode.VAL:
            self._data = datasets.MNIST(self.work_dir, train=False,
                                        transform=transforms.Compose([
                                            transforms.ToTensor(),
                                            transforms.Normalize((0.1307,), (0.3081,))]))
        else:
            raise ValueError("Mode {} not supported yet.".format(mode))

    @property
    def data(self):
        return self._data

    @property
    def size(self):
        return len(self._data)



"""
This module holds classes to model datasets.
"""
from __future__ import absolute_import
import numpy
from six.moves import range


PIXEL_DEPTH = 255


class DataSet(object):

    def __init__(self,
                 images,
                 labels,
                 center=False,
                 scale=False,
                 shuffle=True,
                 fake_data=False):
        if fake_data:
            self._num_examples = 10000
        else:
            assert images.shape[0] == labels.shape[0], (
                "images.shape: %s labels.shape: %s" % (images.shape,
                                                       labels.shape))
            self._num_examples = images.shape[0]
            if center:
                images = images - (PIXEL_DEPTH / 2.0)
            if scale:
                images = images.astype(numpy.float32)
                images = images / PIXEL_DEPTH

        self._images = images
        self._labels = labels
        self.shuffle = shuffle
        self._epochs_completed = 0
        self._index_in_epoch = 0

    @property
    def images(self):
        return self._images

    @property
    def labels(self):
        return self._labels

    @property
    def num_examples(self):
        return self._num_examples

    @property
    def epochs_completed(self):
        return self._epochs_completed

    def next_batch(self, batch_size, fake_data=False):
        """Return the next `batch_size` examples from this data set."""
        if fake_data:
            fake_image = [1.0 for _ in range(784)]
            fake_label = 0
            return [fake_image for _ in range(batch_size)], [
                fake_label for _ in range(batch_size)]
        start = self._index_in_epoch
        self._index_in_epoch += batch_size
        if self._index_in_epoch > self._num_examples:
            # Finished epoch
            self._epochs_completed += 1
            # Shuffle the data
            if self.shuffle:
                perm = numpy.arange(self._num_examples)
                numpy.random.shuffle(perm)
                self._images = self._images[perm]
                self._labels = self._labels[perm]
            # Start next epoch
            start = 0
            self._index_in_epoch = batch_size
            assert batch_size <= self._num_examples
        end = self._index_in_epoch
        return self._images[start:end], self._labels[start:end]


class DataSets(object):
    """
    An class to model common aggregation of datasets.
    """
    def __init__(self, training, test=None, validation=None):
        """
        Create an aggregation of datasets.

        Args:
            training, test, validation: DataSet
        """
        self.training = training
        self.test = test
        self.validation = validation



"""
The module to holds variable initializers.

To get an initializer, call `get`::

    init_para={"stddev": 0.1},
    init = initializers.get(name, **init_para)

where the parameter needed by the initializer is passed in as a dict.

Note that all initializer return tensors/ndarray at least of dimension 1. To
get a tensor of dimension 0 --- that's to say to get a scalar ---, use
`ScalarInitializer` explicitly.
"""
from __future__ import division


from __future__ import absolute_import
from __future__ import print_function
import numpy as np
import math

from ..ops.random_ops import msra_initializer
from .common import SEED
from ..utils import glog as log
from .. import backend as A


# Put all available initializers in a dict, so they can be retrieved by name.
inits = {}


def get(name, **kwargs):
    """
    Get the initializer by name. All necessary parameters should be passed in
    through keyword arguments.

    name: str
        The name of the initializer to use.
    """
    try:
        init = inits[name]
    except KeyError as e:
        if A.backend() == A.TORCH and name == "truncated_normal":
            init = inits["normal"]
            log.warn("Used normal initializer instead of truncated_normal. Truncated normal in not support in torch backend yet.")
        else:
            raise KeyError("`{}`. {} initialization method is not supported".format(e.message, name))

    # Make sure required fields exist.
    for p in init.required_fields:
        if p not in kwargs:
            raise KeyError("Required field {} not found in the provided"
                           " initialization parameters, `init_para`. Perhaps"
                           " you have some typos.".format(p))

    # Fill some default values if not given.
    if name != "constant" and "seed" not in kwargs:
        # Use default seed.
        kwargs["seed"] = SEED

    if name == "range_uniform" and A.backend() == A.TF:
        range = kwargs.pop("range")
        kwargs["minval"] = -range
        kwargs["maxval"] = range
    elif name == "uniform_unit_scaling" or name == "msra" and A.backend() == A.TF:
        if "factor" not in kwargs:
            log.info("Key factor is not found in `init_para`. Use 1")
            kwargs["factor"] = 1

    return init.get(**kwargs)


class InitializerRegistry(object):
    def __init__(self,
                 name,
                 obj,
                 message=None,
                 required_fields=(),
                 default_paras={}):
        """
        Args:
            name: str:
                Name of the initializer. The unique identifier.
            obj: whether is an object, which is used for more
                sophisticated initialization.
            message: str
                Usage help message.
            default_paras: dict
                Default parameters if not specified.
            required_fields: tuple
                A tuple of str that specified required field for this initializer.
        """
        self.obj = obj
        self.message = message
        self.name = name
        self.default_paras = default_paras
        self.required_fields = required_fields
        inits[name] = self

    def get(self, **kwargs):
        if self.obj:
            # Create a callable instance that takes parameters and does the
            # actually initialization.
            for k in self.default_paras:
                if k not in kwargs:
                    kwargs[k] = self.default_paras[k]
            return self.obj(**kwargs)
        else:
            return kwargs["value"]

    def help(self):
        print(self.message)


class Initializer(object):
    """
    Top level class for initializers.
    """
    def __init__(self, seed=SEED):
        np.random.seed(seed)

        # A flag to indicate the initializer is native to akid. This flag is
        # necessary is due to the fact that sometimes circular import may
        # arise, so the issubclass check may not applicable.
        self.native = True


    def compute_fan_in(self, shape):
        if A.backend() == A.TF:
            # The weight layout of tf is HWIO.
            v = 1
            for i in shape[:-1]:
                v *= i
            return v
        elif A.backend() == A.TORCH:
            # The weight layout of tf is OIHW.
            v = 1
            for i in shape[1:]:
                v *= i
            return v

    def compute_fan_out(self, shape):
        if A.backend() == A.TF:
            # The weight layout of tf is HWIO.
            return shape[-1]
        elif A.backend() == A.TORCH:
            # The weight layout of tf is OIHW.
            return shape[0]


class FixedStdInitializer(Initializer):
    def __init__(self, stddev, **kwargs):
        super(FixedStdInitializer, self).__init__(**kwargs)
        self.std = stddev

    def __call__(self, shape):
        value = np.random.normal(loc=0, scale=self.std, size=shape)

        return value


class RangeUniformInitializer(Initializer):
    def __init__(self, low=0, high=1, **kwargs):
        super(RangeUniformInitializer, self).__init__(**kwargs)
        self.low = low
        self.high = high

    def __call__(self, shape):
        v = np.random.uniform(low=self.low, high=self.high, size=shape)
        v = v.astype(A.get_np_dtype())
        return v


class ConstantInitializer(Initializer):
    def __init__(self, value=0):
        super(ConstantInitializer, self).__init__()
        self.value = value

    def __call__(self, shape):
        v = np.ones(shape) * self.value
        v = v.astype(A.get_np_dtype())

        return v


class ScalarInitializer(Initializer):
    def __init__(self, value=0, **kwargs):
        super(ScalarInitializer, self).__init__(**kwargs)
        self.value = value

    def __call__(self, shape):
        # The shape parameter is unused. Just check to make sure it is not
        # accidentally (or intentionally) set wrong.
        if shape != 0:
            raise ValueError("The shape of `ScalarInitializer` should be 0. Got {}".format(shape))

        v = np.array(self.value)
        v = v.astype(A.get_np_dtype())

        return v


class AutoInitializer(Initializer):
    def __init__(self, normal_dist=False, fan_out=False, **kwargs):
        """
        The super class for all initializers that do not need parameters.

        Args:
            normal_dist: bool
                Use normal distribution or not.
        """
        super(AutoInitializer, self).__init__(**kwargs)
        self.normal_dist = normal_dist
        self.fan_out = fan_out


class TheOldInitializer(AutoInitializer):
    """
    The old way of doing initialization before even the (Glorot
    paper)(http://proceedings.mlr.press/v9/glorot10a.html).
    """
    def __call__(self, shape):
        n_in = self.compute_fan_in(shape)
        if self.normal_dist:
            if self.fan_out:
                n_out = self.compute_fan_out(shape)
                stdv = math.sqrt(1. / (n_in + n_out))
            else:
                stdv = math.sqrt(1. / 2 / n_in)
            value = np.random.normal(-stdv, stdv, shape)
        else:
            # NOTE: fan out is implemented for now.
            stdv = 1. / math.sqrt(n_in)
            value = np.random.uniform(-stdv, stdv, shape)

        return value




class XavierInitializer(AutoInitializer):
    """
    The initialization method from [Understanding the difficulty of training deep feedforward neural networks](http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf).
    """
    def __call__(self, shape):
        n_in = self.compute_fan_in(shape)
        n_out = self.compute_fan_out(shape)

        if self.normal_dist:
            std = np.sqrt(2./(n_in + n_out))
            v = np.random.normal(loc=0, scale=std, size=shape)
            v = v.astype(A.get_np_dtype())
        else:
            range = np.sqrt(6./(n_in + n_out))
            v = np.random.uniform(-range, range, size=shape)
            v = v.astype(A.get_np_dtype())

        return v


class MSRAInitializer(AutoInitializer):
    """
    Refer to `akid.ops.random_ops.msra_initializer`.
    """
    def __call__(self, shape):
        n_in = self.compute_fan_in(shape)
        if self.fan_out:
            n_out = self.compute_fan_out(shape)
            std = np.sqrt(4 / (n_in + n_out))
        else:
            std = np.sqrt(2 / n_in)
        v = np.random.normal(loc=0, scale=std, size=shape)
        return v


# CAUTION: if only one required field exists, use ("require",) instead of
# ("require") to make it a tuple.

InitializerRegistry("tensor",
            None,
            "Require fields: value (Tensor holds the initial values)",
            ("value",))
InitializerRegistry("xavier",
                    XavierInitializer,
                    XavierInitializer.__doc__)
InitializerRegistry("scalar",
                    ScalarInitializer,
                    ScalarInitializer.__doc__)

if A.backend() == A.TF:
    import tensorflow as tf
    InitializerRegistry("default",
                        tf.uniform_unit_scaling_initializer,
                        "Uniform unit scaling with factor 1.0/(3)**0.5",
                        # The strange factor here is to make variance
                        # `1/sqrt(dim)`. For the meaning of `dim`, see the doc
                        # of `tf.uniform_unit_scaling_initializer`.
                        default_paras={'factor': 1.0/(3)**0.5})
    InitializerRegistry("normal",
                tf.truncated_normal_initializer,
                "Required fields: stddev (Standard deviation). It is another name "
                "for the truncated normal initializer offered in tensorflow",
                ("stddev",))
    InitializerRegistry("truncated_normal",
                tf.truncated_normal_initializer,
                "Required fields: stddev (Standard deviation)",
                ("stddev",))
    InitializerRegistry("uniform_unit_scaling",
                tf.uniform_unit_scaling_initializer,
                "Optional fields: factor")
    InitializerRegistry("range_uniform",
                tf.random_uniform_initializer,
                "Require fields: range (Range of the uniform init)",
                ("range",))
    InitializerRegistry("msra",
                msra_initializer)
    InitializerRegistry("constant",
                tf.constant_initializer,
                "Require fields: value (initial value)",
                ("value",))
elif A.backend() == A.TORCH:
    InitializerRegistry("default",
                        TheOldInitializer,
                        TheOldInitializer.__doc__)
    InitializerRegistry("constant",
                        ConstantInitializer,
                        None,
                        ("value",))
    InitializerRegistry("normal",
                        FixedStdInitializer,
                        FixedStdInitializer.__doc__,
                        ("stddev",))
    InitializerRegistry("truncated_normal",
                        FixedStdInitializer,
                        FixedStdInitializer.__doc__,
                        ("stddev",))
    InitializerRegistry("msra",
                        MSRAInitializer,
                        MSRAInitializer.__doc__)
    InitializerRegistry("range_uniform",
                        RangeUniformInitializer,
                        RangeUniformInitializer.__doc__)



"""
The distributed computing stack is responsible to handle concurrency and
communication between different computing nodes, so the end user only needs to
deal with how to build a power network. All complexity has been hidden in the
class `Engine`. The usage of `Engine` is just to pick and use.

More specifically, `akid` offers built-in data parallel scheme in form of class
`Engine`. Currently, the engine mainly works with neural network training,
which is be used with `Kid` by specifying the engine at the construction of the
kid.

As an example, we could do data parallelism on multiple computing towers using::

    kid = kids.Kid(
        sensor,
        brain,
        MomentumKongFu(lr_scheme={"name": LearningRateScheme.placeholder}),
        engine={"name": "data_parallel", "num_gpu": 2},
        log_dir="log",
        max_epoch=200)
"""
from __future__ import absolute_import
import abc

import tensorflow as tf

from .blocks import ValidatableProcessingBlock
from .interface_blocks import UpdateBlock
from .. import backend as A
from .. import parallel as P
from ..utils.tools import is_tuple_or_list
from six.moves import range
from six.moves import zip


class Engine(ValidatableProcessingBlock, UpdateBlock):
    """
    The class that abstracts parallel scheme of network training.

    An `Engine` is responsible for setting up computational graph of a `Brain`
    on proper devices, and the coordination between devices (if there is any).

    More specifically, an `Engine` will take the data from `Sensor`, which
    means it has already been set up, and set up `Brain` and `KongFu`. It
    splits the data provide by `Sensor`, feeds them to devices according to the
    parallel scheme, gathers and processes the results, and provides the end
    result as if no parallelism exists.
    """
    def __init__(self, brain, kongfu, **kwargs):
        if "name" not in kwargs:
            kwargs["name"] = 'engine'
        kwargs["call_on_update"] = True

        super(Engine, self).__init__(**kwargs)

        self.brain = brain
        self.kongfu = kongfu

    @abc.abstractmethod
    def loss(self, get_val=False):
        """
        Return the loss of a `Brain`. If `get_val` is True, return the
        validation loss, otherwise, return training loss.
        """
        raise NotImplementedError("Each engine should implement the interface"
                                  " to provide loss.")

    @abc.abstractmethod
    def eval(self, get_val=False):
        """
        Return the evaluation list of a `Brain`. If `get_val` is True, return
        that of validation brain, otherwise, that of training brain.
        """
        raise NotImplementedError("Each engine should implement the interface"
                                  " to provide evaluation.")

    @property
    def data(self):
        """
        Placeholder. Do not return anything. Ideally, it should return whatever
        data the brain it contains. But it does not of use for now. So just
        skip.
        """
        return None

    # def _post_forward(self):
    #     pass
        # self.train_op_list.append(apply_grad_op)
        # train_op = tf.group(*self.train_op_list)

        # with tf.control_dependencies([train_op]):
        #     update_op = self.brain.on_para_update()
        #     if update_op:
        #         self.train_op = tf.group(*update_op)
        #     else:
        #         self.train_op = train_op

        # for grad, var in self.grads:
        #     if grad is not None:
        #         tf.summary.histogram(
        #             var.op.name + '/gradients',
        #             grad,
        #             collections=[TRAINING_DYNAMICS_COLLECTION])


class SingleGPUEngine(Engine):
    def _setup(self):
        if not self.brain.is_setup:
            self.brain.setup()
        if A.backend() == A.TF:
            self.val_brain = self.brain.get_val_copy()
        elif A.backend() == A.TORCH:
            self.val_brain = self.brain

        self.kongfu.set_var_list(self.brain.get_filters())
        self.kongfu.setup()

    def _forward(self, data, val=False):
        if val:
            if A.backend() == A.TORCH:
                # For torch, val brain and brain are the same.
                pushed_brain_state = None
                if self.val_brain.mode != A.Mode.VAL:
                    pushed_brain_state = self.val_brain.mode
                    self.val_brain.set_val(True)
                pushed_name = self.brain.name
                self.val_brain.name = self.brain.name + "_val"
                d = self.val_brain.forward(data)
                if pushed_brain_state is not None:
                    # Right now, we only have train and val for a brain, thus
                    # the saved state is not used actually.
                    self.val_brain.set_val(False)
                self.val_brain.name = pushed_name
            elif A.backend() == A.TF:
                d = self.val_brain.forward(data)
            else:
                raise ValueError("Backend not supported.")
        else:
            pushed_brain_state = None
            if self.brain.mode != A.Mode.TRAIN:
                pushed_brain_state = self.brain.mode
                self.brain.set_val(False)

            d = self.brain.forward(data)

            if pushed_brain_state != None:
                self.brain.set_val(True)
        return d

    def _update(self):
        grads = self.kongfu.forward(self.brain.loss)
        self.train_op = self.kongfu.update(grads)
        return self.train_op

    def loss(self, get_val=False):
        if not get_val:
            return self.brain.loss
        else:
            return self.val_brain.loss

    def eval(self, get_val=False):
        if not get_val:
            return self.brain.eval
        else:
            return self.val_brain.eval

    def verbose_eval(self, get_val=False):
        if not get_val:
            return self.brain.verbose_eval
        else:
            return self.val_brain.verbose_eval

    def get_layer_data(self, name, get_val=False):
        """
        Given the `name` of the layer, return the tensor of the data of this
        layer. `get_val` has similar meaning with `loss`.
        """
        if get_val:
            brain = self.val_brain
        else:
            brain = self.brain
        layer = None
        for b in brain.blocks:
            if b.name == name:
                layer = b
                break
        if layer is None:
            raise Exception("Layer {} is not found.".format(name))

        return layer.data

    def get_layer(self, name, get_val=False):
        """
        Given the `name` of the layer, return the tensor of the data of this
        layer. `get_val` has similar meaning with `loss`.
        """
        if get_val:
            brain = self.val_brain
        else:
            brain = self.brain

        layer = None
        for b in brain.blocks:
            if b.name == name:
                layer = b
                break
        if layer is None:
            raise Exception("Layer {} is not found.".format(name))

        return layer


class DataParallelEngine(SingleGPUEngine):
    """
    This engine will implement typical parallelism in training neural
    network. It splits the batch, and train a fraction of them in an individual
    computing devices.

    Due to the known fact that communication between GPUs are slow, the average
    of gradient is done on CPU.
    """
    def __init__(self, gpu_num=2, val_gpu_num=None, **kwargs):
        super(DataParallelEngine, self).__init__(**kwargs)
        # TODO: if num_gpu is None, use all available ones.
        self.train_gpu_num = gpu_num
        self.train_devices = [i for i in range(gpu_num)]
        if val_gpu_num is not None:
            self.val_gpu_num = val_gpu_num
            self.val_devices = [i for i in range(val_gpu_num)]
        else:
            self.val_gpu_num = gpu_num
            self.val_devices = self.train_devices

        self.gpu_num = self.train_gpu_num
        self.devices = self.train_devices

        if A.backend() != A.TORCH:
            raise ValueError("Backend other torch has not been implemented yet.")
            # TODO: note that the weight decay computation in the old code of
            # tensorflow may be wrong, which makes each replica creates a copy
            # of the weight decay. It may make the weight decay 8 or 4 larger
            # than the reported value, which may be the reason why the
            # reproduced resnet has a lower accuracy. Check when implement
            # tensorflow data parallel.

    def set_val(self, val):
        super(DataParallelEngine, self).set_val(val)

        if val:
            self.devices = self.val_devices
            self.gpu_num = self.val_gpu_num
        else:
            self.devices = self.train_devices
            self.gpu_num = self.train_gpu_num

    def _setup(self):
        if self.gpu_num == 1:
            super(DataParallelEngine, self)._setup()
            return

        # Set up the primary computational graph, and replicate.
        if not self.brain.is_setup:
            self.brain.setup()
        A.get_variable_scope().reuse_variables()
        # Replicate the parameters.
        self._sync(self.brain.get_filters(), self.devices)

        self.towers = self._shadow_copy(self.brain)
        if A.backend() == A.TF:
            self.val_brain = self.brain.get_val_copy()
            self.val_towers = self._shadow_copy(self.val_brain)
        else:
            self.val_brain = self.brain
            self.val_towers = self.towers

        self.kongfu.set_var_list(self.brain.get_filters())
        self.kongfu.setup()

    def _shadow_copy(self, brain):
        """
        Make `self.gpu_num` number of shadow copy of `brain`. A list of shadow
        copies will be returned. The original brain will be put at the first of
        the list.
        """
        # Set up shadow copies.
        towers = []
        towers.append(brain)
        for i in range(1, self.gpu_num):
            with P.device('/gpu:{}'.format(i)):
                tower = brain.get_shadow_copy()
                tower.setup()
                towers.append(tower)

        return towers

    def _forward(self, data, val=False):
        if self.gpu_num == 1:
            d = super(DataParallelEngine, self)._forward(data, val)
            return d

        # Else, do data parallel.
        # Scatter the data to shadow devices.
        data = P.scatter(data, self.devices)

        # Forward propagate the scattered data.
        if A.backend() == A.TORCH:
            if val:
                pushed_name = self.brain.name
                for t in self.towers:
                    t.set_val(True)
                    t.name = t.name + "_val"

        # results here is a general mechanism to gather result, which is
        # another use here though, since loss and evaluation metrics is
        # gathered through loss and eval in towers' attributes.
        results = []
        threads = []
        for i in range(self.gpu_num):
            with P.device("/gpu:{}".format(i)):
                # TODO: think how to handle tensorflow
                ret, t = P.thread_run(self.towers[i].forward, data[i])
                results.append(ret)
                threads.append(t)
        for t in threads:
            t.join()
        # Gather the result, and compute average loss
        losses = [t.loss for t in self.towers]

        loss = self._average(losses)
        evals = [t.eval for t in self.towers]
        eval = self._average(evals)
        # To keep return values, a list is created in `thread_run` to get the
        # results. It only has one element, so we retrieve it.
        results = [r[0] for r in results]
        if None not in results:
            ret = self._average(results)
        else:
            ret = None

        if val:
            self._val_loss = loss
            self._val_eval = eval
            for t in self.towers:
                t.set_val(False)
                t.name = pushed_name
        else:
            self._train_loss = loss
            self._train_eval = eval

        return ret

    def _update(self, *args, **kwargs):
        # TODO: test the behavior of tensorflow, given that the loss is gather
        # from different GPU
        grads = self.kongfu.forward(self._train_loss)
        self.train_op = self.kongfu.update(grads)
        return self.train_op

    def _sync(self, variables, devices):
        """
        Synchronize parameters to devices specified.

        Args:
            variables: a list of variables
            devices: a list of id for GPUs.
        """
        if A.backend() == A.TORCH:
            # Tensorflow handles all communication, no need for this. Do this
            # for torch.
            P.broadcast(variables, devices)

    def _on_update(self, *args, **kwargs):
        # If the backend is torch, broadcast the parameters from primary device
        # to shadow copies.
        if A.backend() == A.TORCH:
            self._sync(self.brain.get_filters(), self.devices)
            # Let the shadow copy retried its parameters again.
            with A.variable_scope(self.name):
                A.get_variable_scope().reuse_variables()
                self.towers = self._shadow_copy(self.brain)

    def get_layer_data(self, name, get_val=False):
        if get_val:
            towers = self.val_towers
        else:
            towers = self.train_towers

        data = []
        for t in towers:
            l = t.get_layer_by_name(name)
            data.append(l.data)

        # Check whether the output is a list, if it is merge by list members
        if type(data[0]) is list:
            data = list(zip(*data))
            for i, l in enumerate(data):
                data[i] = tf.concat(axis=0, values=l, name=name)
        # Otherwise, concat.
        else:
            data = tf.concat(axis=0, values=data, name=name)

        return data

    def _split_input(self, data, label):
        """
        Given data and labels, split them and return.
        """
        with A.variable_scope("data_split"):
            splitted_data = tf.split(axis=0, num_or_size_splits=self.gpu_num, value=data)
            if type(label) is list:
                splitted_labels = []
                for i in range(0, len(label)):
                    splitted_labels.append(tf.split(axis=0, num_or_size_splits=self.gpu_num, value=label[i]))
                splitted_labels = list(zip(*splitted_labels))
            else:
                splitted_labels = tf.split(axis=0, num_or_size_splits=self.gpu_num, value=label)

        return splitted_data, splitted_labels

    def _average(self, data):
        """
        Given a list of computing towers, average their loss, and return.
        """
        # TODO: think how the gather works in tensorflow
        data_gathered = P.gather(data)
        def mean_inner(d, name=None):
            if is_tuple_or_list(d):
                return list(map(mean_inner, d))
            else:
                d_reduced = A.mean(d)
                return d_reduced

        data_reduced = mean_inner(data_gathered)

        if is_tuple_or_list(data_reduced):
            # Rename all the results. The results are a list of tensors, or named
            # tuples. So an iteration would do.
            for i, r in enumerate(data_reduced):
                name = A.get_name(data[0][i], with_device_id=False)
                if name:
                    new_name = name # NOTE: Use the same name for now.
                    if A.is_tensor(r):
                        A.cache_tensor_auto_scope(r, new_name)
                    elif isinstance(data[0][i], A.NamedValue):
                        r = type(data[0][i])(new_name, r)
                        # TODO: cache the tuple results separately
                        data_reduced[i] = r
        else:
            name = A.get_name(data[0])
            if name:
                new_name = name
                A.cache_tensor_auto_scope(data_reduced, new_name)

        return data_reduced

    def _average_eval(self, towers):
        """
        Given a list of computing towers, average their evaluation metrics and
        return.
        """
        with A.variable_scope("eval_average"):
            eval_list = []
            for i in range(0, len(towers[0].eval)):
                sum_of_eval = tf.add_n([t.eval[i] for t in towers])
                eval = tf.div(sum_of_eval,
                              self.gpu_num,
                              name="{}_avg".format(towers[0].eval[i].op.name))
                eval_list.append(eval)

        return eval_list

    def _forward_train(self):
        # Split the data.
        data = self.sensor.data()
        label = self.sensor.labels()

        # TODO: A description of how the data parallelism is done should be
        # provided. Otherwise, the logging information only has the one of the
        # first built tower.

        splitted_data, splitted_labels = self._split_input(data, label)

        # Set up brains according to the number of gpus used.

        # Since we need reduce operations on replicas of brains, we need to
        # keep track of different brains, denoted as towers of the
        # brain. Similar with validation towers.
        self.train_towers = []
        tower_grads = []
        tower = self.brain
        kongfu = self.kongfu
        self.train_op_list = []
        for i in range(0, self.gpu_num):
            self.log("Setting up tower {} for training".format(i))
            with tf.device('/gpu:{}'.format(i)):
                # Set up a tower
                system_in = self._setup_system_in(splitted_data[i],
                                                  splitted_labels[i])
                tower.forward(system_in)
                if type(tower.train_op) is list:
                     self.train_op_list.extend(tower.train_op)
                else:
                    self.train_op_list.append(self.brain.train_op)

                # Keep track of the new tower.
                self.train_towers.append(tower)

                # Set up KongFu (optimizer).
                # For now, we do not need to keep track of Kongfu, so just set
                # it up multiple times.
                kongfu.forward(tower.loss)

                # Create the next tower.
                # Do not do copy at the last tower.
                if i is not self.gpu_num - 1:
                    tower = tower.get_shadow_copy()
                    kongfu = kongfu.get_shadow_copy()

                tower_grads.append(kongfu.data)

        # Gather and reduce.
        with tf.device('/cpu:0'):
            self.grads = self._average_grads(tower_grads)
            self._train_loss = self._average_loss(self.train_towers)
            self._train_eval = self._average_eval(self.train_towers)

        return self.grads

    def _forward_val(self):
        data = self.sensor.data(get_val=True)
        label = self.sensor.labels(get_val=True)
        splitted_data, splitted_labels = self._split_input(data, label)

        # Set up val brains according to the number of gpus used.
        self.val_towers = []
        tower = self.val_brain
        for i in range(0, self.gpu_num):
            self.log("Setting up tower {} for validation".format(i))
            with tf.device('/gpu:{}'.format(i)):
                system_in = self._setup_system_in(splitted_data[i],
                                                  splitted_labels[i])
                tower.forward(system_in)
                self.val_towers.append(tower)
                if i is not self.gpu_num - 1:
                    tower = tower.get_shadow_copy()

        with tf.device('/cpu:0'):
            self._val_loss = self._average_loss(self.val_towers)
            self._val_eval = self._average_eval(self.val_towers)

    def loss(self, get_val=False):
        if not get_val:
            return self._train_loss
        else:
            return self._val_loss

    def eval(self, get_val=False):
        if not get_val:
            return self._train_eval
        else:
            return self._val_eval

    def _average_grads(self, tower_grads):
        """
        Calculate the average gradient for each shared variable across all
        towers.

        Note that this function provides a synchronization point across all
        towers.

        Args:
            tower_grads: List of lists of (gradient, variable) tuples.
                The outer list is over individual gradients. The inner list is
                over the gradient calculation for each tower.
        Returns:
            List of pairs of (gradient, variable), where the gradient has been
            averaged across all towers.
        """
        with A.variable_scope("gradient_average"):
            average_grads = []
            for grad_and_vars in zip(*tower_grads):
                # Note that each grad_and_vars looks like the following:
                #   ((grad0_gpu0, var0_gpu0), ... , (grad0_gpuN, var0_gpuN))
                grads = []
                for g, _ in grad_and_vars:
                    # Add 0 dimension to the gradients to represent the tower.
                    expanded_g = tf.expand_dims(g, 0)

                    # Append on a 'tower' dimension which we will average over
                    # below.
                    grads.append(expanded_g)

                # Average over the 'tower' dimension.
                grad = tf.concat(axis=0, values=grads)
                grad = tf.reduce_mean(grad, 0)

                # Keep in mind that the Variables are redundant because they
                # are shared across towers. So .. we will just return the first
                # tower's pointer to the Variable.
                v = grad_and_vars[0][1]
                grad_and_var = (grad, v)
                average_grads.append(grad_and_var)

        return average_grads


engines = {}


class EngineRegistry(object):
    def __init__(self,
                 name,
                 cls,
                 message=None,
                 required_fields=(),
                 default_paras={}):
        self.message = message
        self.cls = cls
        self.name = name
        self.default_paras = default_paras
        self.required_fields = required_fields
        engines[name] = self


def get(name, **kwargs):
    entry = engines[name]
    for f in entry.required_fields:
        if f not in kwargs:
            raise KeyError("Required field {} not found in the engine parameters.".format(entry.required_fields))

    for k in entry.default_paras:
        if k not in kwargs:
            kwargs[k] = entry.default_paras[k]

    return engines[name].cls(**kwargs)


EngineRegistry('single',
               SingleGPUEngine,
               "Single GPU engine")
EngineRegistry('data_parallel',
               DataParallelEngine,
               "Data Parallel Engine. Specify the number of GPU with field `gpu_num` (2 GPU is used by default).",
               default_paras={"gpu_num": 2})



"""
The module provides samplers to iterate data in different sequence.
"""
from __future__ import print_function
from __future__ import absolute_import
from six import raise_from

import random

from ..utils import glog as log
from .events import EpochCompletedEvent
from six.moves import range


sampler_dict = {}


def get(sampler_name, *args, **kwargs):
    """
    Given the name of the sampler, return a Sampler that has a `next()`
    interface to iterate the dataset.
    """
    try:
        sampler = sampler_dict[sampler_name]
    except ValueError as e:
        raise_from(ValueError("Sampler {} is not supported".format(sampler_name)), e)

    return sampler.get(*args, **kwargs)


class SamplerRegistry(object):
    def __init__(self,
                 name,
                 class_name,
                 message=None):
        """
        Args:
            name: str:
                Name of the initializer. The unique identifier.
            class_name: class
                A Sampler class.
            message: str
                Usage help message.
        """
        self.message = message
        self.name = name
        self.class_name = class_name
        sampler_dict[name] = self

    def get(self, *args, **kwargs):
        return self.class_name(*args, **kwargs)

    def help(self):
        print(self.message)


class Sampler(object):
    """
    Given a list of indices, a `Sampler` provides a function `next` to return
    the next batch of indices in a certain order.
    """
    def __init__(self, length, *args, **kwargs):
        super(Sampler, self).__init__(*args, **kwargs)
        self.indices = list(range(length))
        self.current_idx = 0

    def reset(self):
        self.current_idx = 0

    def _check_epoch_finishes(self):
        if self.current_idx >= len(self.indices):
            self.reset()
            log.debug("An epoch finished. Reset.")
            raise EpochCompletedEvent()

class SequenceSampler(Sampler):
    def next(self, size=1):
        self._check_epoch_finishes()

        if self.current_idx + size > len(self.indices):
            index_batch = self.indices[self.current_idx:]
            self.current_idx = len(self.indices)
        else:
            index_batch = self.indices[self.current_idx:self.current_idx+size]
            self.current_idx += size

        return index_batch


class ShuffleSampler(SequenceSampler):
    def __init__(self, *args, **kwargs):
        super(ShuffleSampler, self).__init__(*args, **kwargs)
        random.shuffle(self.indices)

    def reset(self):
        super(ShuffleSampler, self).reset()
        random.shuffle(self.indices)


SamplerRegistry("shuffle",
                ShuffleSampler,
                message="Randomly shuffle the dataset for each epoch.")
SamplerRegistry("sequence",
                SequenceSampler,
                message="Sequentially sample the dataset for each epoch.")



"""Functions for downloading and reading MNIST data."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gzip
import os
import six.moves.urllib.request, six.moves.urllib.parse, six.moves.urllib.error

import numpy
from six.moves import xrange  # pylint: disable=redefined-builtin
from six.moves import range

SOURCE_URL = 'http://yann.lecun.com/exdb/mnist/'


def maybe_download(filename, work_directory):
  """Download the data from Yann's website, unless it's already here."""
  if not os.path.exists(work_directory):
    os.mkdir(work_directory)
  filepath = os.path.join(work_directory, filename)
  if not os.path.exists(filepath):
    filepath, _ = six.moves.urllib.request.urlretrieve(SOURCE_URL + filename, filepath)
    statinfo = os.stat(filepath)
    print('Succesfully downloaded', filename, statinfo.st_size, 'bytes.')
  return filepath


def _read32(bytestream):
  dt = numpy.dtype(numpy.uint32).newbyteorder('>')
  return numpy.frombuffer(bytestream.read(4), dtype=dt)


def extract_images(filename):
  """Extract the images into a 4D uint8 numpy array [index, y, x, depth]."""
  print('Extracting', filename)
  with gzip.open(filename) as bytestream:
    magic = _read32(bytestream)
    if magic != 2051:
      raise ValueError(
          'Invalid magic number %d in MNIST image file: %s' %
          (magic, filename))
    num_images = _read32(bytestream)
    rows = _read32(bytestream)
    cols = _read32(bytestream)
    buf = bytestream.read(rows * cols * num_images)
    data = numpy.frombuffer(buf, dtype=numpy.uint8)
    data = data.reshape(num_images, rows, cols, 1)
    return data


def dense_to_one_hot(labels_dense, num_classes=10):
  """Convert class labels from scalars to one-hot vectors."""
  num_labels = labels_dense.shape[0]
  index_offset = numpy.arange(num_labels) * num_classes
  labels_one_hot = numpy.zeros((num_labels, num_classes))
  labels_one_hot.flat[index_offset + labels_dense.ravel()] = 1
  return labels_one_hot


def extract_labels(filename, one_hot=False):
  """Extract the labels into a 1D uint8 numpy array [index]."""
  print('Extracting', filename)
  with gzip.open(filename) as bytestream:
    magic = _read32(bytestream)
    if magic != 2049:
      raise ValueError(
          'Invalid magic number %d in MNIST label file: %s' %
          (magic, filename))
    num_items = _read32(bytestream)
    buf = bytestream.read(num_items)
    labels = numpy.frombuffer(buf, dtype=numpy.uint8)
    if one_hot:
      return dense_to_one_hot(labels)
    return labels


class DataSet(object):

  def __init__(self, images, labels, fake_data=False):
    if fake_data:
      self._num_examples = 10000
    else:
      assert images.shape[0] == labels.shape[0], (
          "images.shape: %s labels.shape: %s" % (images.shape,
                                                 labels.shape))
      self._num_examples = images.shape[0]

      # Convert shape from [num examples, rows, columns, depth]
      # to [num examples, rows*columns] (assuming depth == 1)
      assert images.shape[3] == 1
      # images = images.reshape(images.shape[0],
      #                         images.shape[1] * images.shape[2])
      # Convert from [0, 255] -> [0.0, 1.0].
      images = images.astype(numpy.float32)
      images = numpy.multiply(images, 1.0 / 255.0)
    self._images = images
    self._labels = labels
    self._epochs_completed = 0
    self._index_in_epoch = 0

  @property
  def images(self):
    return self._images

  @property
  def labels(self):
    return self._labels

  @property
  def num_examples(self):
    return self._num_examples

  @property
  def epochs_completed(self):
    return self._epochs_completed

  def next_batch(self, batch_size, fake_data=False):
    """Return the next `batch_size` examples from this data set."""
    if fake_data:
      fake_image = [1.0 for _ in range(784)]
      fake_label = 0
      return [fake_image for _ in range(batch_size)], [
          fake_label for _ in range(batch_size)]
    start = self._index_in_epoch
    self._index_in_epoch += batch_size
    if self._index_in_epoch > self._num_examples:
      # Finished epoch
      self._epochs_completed += 1
      # Shuffle the data
      perm = numpy.arange(self._num_examples)
      numpy.random.shuffle(perm)
      self._images = self._images[perm]
      self._labels = self._labels[perm]
      # Start next epoch
      start = 0
      self._index_in_epoch = batch_size
      assert batch_size <= self._num_examples
    end = self._index_in_epoch
    return self._images[start:end], self._labels[start:end]


def read_data_sets(train_dir, fake_data=False, one_hot=False):
  class DataSets(object):
    pass
  data_sets = DataSets()

  if fake_data:
    data_sets.train = DataSet([], [], fake_data=True)
    data_sets.validation = DataSet([], [], fake_data=True)
    data_sets.test = DataSet([], [], fake_data=True)
    return data_sets

  TRAIN_IMAGES = 'train-images-idx3-ubyte.gz'
  TRAIN_LABELS = 'train-labels-idx1-ubyte.gz'
  TEST_IMAGES = 't10k-images-idx3-ubyte.gz'
  TEST_LABELS = 't10k-labels-idx1-ubyte.gz'
  VALIDATION_SIZE = 5000

  local_file = maybe_download(TRAIN_IMAGES, train_dir)
  train_images = extract_images(local_file)

  local_file = maybe_download(TRAIN_LABELS, train_dir)
  train_labels = extract_labels(local_file, one_hot=one_hot)

  local_file = maybe_download(TEST_IMAGES, train_dir)
  test_images = extract_images(local_file)

  local_file = maybe_download(TEST_LABELS, train_dir)
  test_labels = extract_labels(local_file, one_hot=one_hot)

  validation_images = train_images[:VALIDATION_SIZE]
  validation_labels = train_labels[:VALIDATION_SIZE]
  train_images = train_images[VALIDATION_SIZE:]
  train_labels = train_labels[VALIDATION_SIZE:]

  data_sets.train = DataSet(train_images, train_labels)
  data_sets.validation = DataSet(validation_images, validation_labels)
  data_sets.test = DataSet(test_images, test_labels)

  return data_sets



from __future__ import absolute_import
from akid import AKID_DATA_PATH
from akid import FeedSensor
from akid import Kid
from akid import MomentumKongFu
from akid import MNISTFeedSource

from akid.models.brains import LeNet

brain = LeNet(name="LeNet")
source = MNISTFeedSource(name="MNIST",
                         url='http://yann.lecun.com/exdb/mnist/',
                         work_dir=AKID_DATA_PATH + '/mnist',
                         center=True,
                         scale=True,
                         num_train=60000,
                         num_val=10000)

s = Kid(FeedSensor(name='data', source_in=source),
        brain,
        MomentumKongFu(),
        max_steps=1000)
s.setup()

s.practice()



"""Trains and Evaluates the MNIST network using a feed dictionary.

TensorFlow install instructions:
https://tensorflow.org/get_started/os_setup.html

MNIST tutorial:
https://tensorflow.org/tutorials/mnist/tf/index.html

"""
# pylint: disable=missing-docstring
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os.path
import time

import tensorflow.python.platform
import numpy
from six.moves import xrange  # pylint: disable=redefined-builtin
import tensorflow as tf

import input_data
import mnist
from six.moves import range


# Basic model parameters as external flags.
flags = tf.app.flags
FLAGS = flags.FLAGS
flags.DEFINE_float('learning_rate', 0.01, 'Initial learning rate.')
flags.DEFINE_integer('max_steps', 8000, 'Number of steps to run trainer.')
flags.DEFINE_integer('batch_size', 100, 'Batch size.  '
                     'Must divide evenly into the dataset sizes.')
flags.DEFINE_string('train_dir', 'data', 'Directory to put the training data.')
flags.DEFINE_string('data_dir', 'data', 'Directory to put the training data.')
flags.DEFINE_boolean('fake_data', False, 'If true, uses fake data '
                     'for unit testing.')


def placeholder_inputs(batch_size):
    """Generate placeholder variables to represent the the input tensors.

    These placeholders are used as inputs by the rest of the model building
    code and will be fed from the downloaded data in the .run() loop, below.

    Args:
        batch_size: The batch size will be baked into both placeholders.

    Returns:
        images_placeholder: Images placeholder.
        labels_placeholder: Labels placeholder.
    """
    # Note that the shapes of the placeholders match the shapes of the full
    # image and label tensors, except the first dimension is now batch_size
    # rather than the full size of the train or test data sets.
    images_placeholder = tf.placeholder(tf.float32, shape=(batch_size,
                                                           mnist.IMAGE_SIZE,
                                                           mnist.IMAGE_SIZE,
                                                           1))
    labels_placeholder = tf.placeholder(tf.int32, shape=(batch_size))
    return images_placeholder, labels_placeholder


def fill_feed_dict(data_set, images_pl, labels_pl):
    """Fills the feed_dict for training the given step.

    A feed_dict takes the form of:
    feed_dict = {
        <placeholder>: <tensor of values to be passed for placeholder>,
        ....
    }

    Args:
        data_set: The set of images and labels, from
                  input_data.read_data_sets()
        images_pl: The images placeholder, from placeholder_inputs().
        labels_pl: The labels placeholder, from placeholder_inputs().

    Returns:
        feed_dict: The feed dictionary mapping from placeholders to values.
    """
    # Create the feed_dict for the placeholders filled with the next
    # `batch size ` examples.
    images_feed, labels_feed = data_set.next_batch(FLAGS.batch_size,
                                                   FLAGS.fake_data)
    feed_dict = {
        images_pl: images_feed,
        labels_pl: labels_feed,
    }
    return feed_dict


def do_eval(sess,
            eval_correct,
            images_placeholder,
            labels_placeholder,
            data_set):
    """Runs one evaluation against the full epoch of data.

    Args:
        sess: The session in which the model has been trained.
        eval_correct: The Tensor that returns the number of correct
             predictions.
        images_placeholder: The images placeholder.
        labels_placeholder: The labels placeholder.
        data_set: The set of images and labels to evaluate, from
        input_data.read_data_sets().
    """
    # And run one epoch of eval.
    true_count = 0  # Counts the number of correct predictions.
    steps_per_epoch = data_set.num_examples // FLAGS.batch_size
    num_examples = steps_per_epoch * FLAGS.batch_size
    for step in range(steps_per_epoch):
        feed_dict = fill_feed_dict(data_set,
                                   images_placeholder,
                                   labels_placeholder)
        true_count += sess.run(eval_correct, feed_dict=feed_dict)
    precision = true_count / num_examples
    print('  Num examples: %d  Num correct: %d  Precision @ 1: %0.04f' %
          (num_examples, true_count, precision))


def run_training():
    """Train MNIST for a number of steps."""
    # Get the sets of images and labels for training, validation, and
    # test on MNIST.
    data_sets = input_data.read_data_sets(FLAGS.data_dir, FLAGS.fake_data)

    # Tell TensorFlow that the model will be built into the default Graph.
    with tf.Graph().as_default():
        # Generate placeholders for the images and labels.
        images_placeholder, labels_placeholder = placeholder_inputs(
            FLAGS.batch_size)

        # Build a Graph that computes predictions from the inference model.
        logits = mnist.inference(images_placeholder)

        # Add to the Graph the Ops for loss calculation.
        loss = mnist.loss(logits, labels_placeholder)

        # Add to the Graph the Ops that calculate and apply gradients.
        train_op = mnist.training(loss, FLAGS.batch_size)

        # Add the Op to compare the logits to the labels during evaluation.
        eval_correct = mnist.evaluation(logits, labels_placeholder)

        # Build the summary operation based on the TF collection of Summaries.
        summary_op = tf.summary.merge_all()

        # Create a saver for writing training checkpoints.
        saver = tf.train.Saver()

        # Create a session for running Ops on the Graph.
        sess = tf.Session()

        # Run the Op to initialize the variables.
        init = tf.global_variables_initializer()
        sess.run(init)

        # Instantiate a SummaryWriter to output summaries and the Graph.
        summary_writer = tf.summary.FileWriter(FLAGS.train_dir,
                                                graph_def=sess.graph_def)

        # And then after everything is built, start the training loop.
        for step in range(FLAGS.max_steps):
            start_time = time.time()

            # Fill a feed dictionary with the actual set of images and labels
            # for this particular training step.
            feed_dict = fill_feed_dict(data_sets.train,
                                       images_placeholder,
                                       labels_placeholder)

            # Run one step of the model.  The return values are the activations
            # from the `train_op` (which is discarded) and the `loss` Op.  To
            # inspect the values of your Ops or variables, you may include them
            # in the list passed to sess.run() and the value tensors will be
            # returned in the tuple from the call.
            _, loss_value = sess.run([train_op, loss],
                                     feed_dict=feed_dict)

            duration = time.time() - start_time

            # Write the summaries and print an overview fairly often.
            if step % 100 == 0:
                # Print status to stdout.
                print('Step %d: loss = %.2f (%.3f sec)'
                      % (step, loss_value, duration))
                # Update the events file.
                summary_str = sess.run(summary_op, feed_dict=feed_dict)
                summary_writer.add_summary(summary_str, step)

            # Save a checkpoint and evaluate the model periodically.
            if (step + 1) % 1000 == 0 or (step + 1) == FLAGS.max_steps:
                saver.save(sess, FLAGS.train_dir, global_step=step)
                # Evaluate against the training set.
                print('Training Data Eval:')
                do_eval(sess,
                        eval_correct,
                        images_placeholder,
                        labels_placeholder,
                        data_sets.train)
                # Evaluate against the validation set.
                print('Validation Data Eval:')
                do_eval(sess,
                        eval_correct,
                        images_placeholder,
                        labels_placeholder,
                        data_sets.validation)
                # Evaluate against the test set.
                print('Test Data Eval:')
                do_eval(sess,
                        eval_correct,
                        images_placeholder,
                        labels_placeholder,
                        data_sets.test)


def main(_):
    run_training()


if __name__ == '__main__':
    tf.app.run()



from __future__ import absolute_import
from akid.train.tuner import tune


net_paras_list = []
net_paras_list.append({
    "activation": [
        {"type": "relu"},
        {"type": "relu"},
        {"type": "relu"},
        {"type": "relu"}],
    "bn": True})
net_paras_list.append({
    "activation": [
        {"type": "maxout", "group_size": 2},
        {"type": "maxout", "group_size": 2},
        {"type": "maxout", "group_size": 2},
        {"type": "maxout", "group_size": 5}],
    "bn": True})
net_paras_list.append({
    "activation": [
        {"type": "gsoftmax", "group_size": 2},
        {"type": "gsoftmax", "group_size": 2},
        {"type": "gsoftmax", "group_size": 2},
        {"type": "gsoftmax", "group_size": 5}],
    "bn": True})

opt_paras_list = []
opt_paras_list.append({"lr": 0.025})
opt_paras_list.append({"lr": 0.05})


def setup(graph):
    from akid import AKID_DATA_PATH
    from akid import GraphBrain, Cifar10FeedSource, FeedSensor, Kid
    from akid import MomentumKongFu
    from akid.layers import DropoutLayer
    from akid.sugar import cnn_block

    brain = GraphBrain(name="maxout")

    brain.attach(DropoutLayer(keep_prob=0.8, name='dropout0'))

    brain.attach(cnn_block(
        ksize=[8, 8],
        init_para={
            "name": "uniform",
            "range": 0.005},
        wd={"type": "l2", "scale": 0.0005},
        out_channel_num=192,
        pool_size=[4, 4],
        pool_stride=[2, 2],
        activation={{ net_paras["activation"][0] }},
        keep_prob=0.5,
        bn={{ net_paras["bn"] }}))

    brain.attach(cnn_block(
        ksize=[8, 8],
        init_para={
            "name": "uniform",
            "range": 0.005},
        wd={"type": "l2", "scale": 0.0005},
        out_channel_num=384,
        pool_size=[4, 4],
        pool_stride=[2, 2],
        activation={{ net_paras["activation"][1] }},
        keep_prob=0.5,
        bn={{ net_paras["bn"] }}))

    brain.attach(cnn_block(
        ksize=[5, 5],
        init_para={
            "name": "uniform",
            "range": 0.005},
        wd={"type": "l2", "scale": 0.0005},
        out_channel_num=384,
        pool_size=[2, 2],
        pool_stride=[2, 2],
        activation={{ net_paras["activation"][2] }},
        keep_prob=0.5,
        bn={{ net_paras["bn"] }}))

    brain.attach(cnn_block(
        init_para={
            "name": "uniform",
            "range": 0.005},
        wd={"type": "l2", "scale": 0.004},
        out_channel_num=2500,
        activation={{ net_paras["activation"][3] }},
        keep_prob=0.5,
        bn={{ net_paras["bn"] }}))

    brain.attach(cnn_block(
        init_para={
            "name": "uniform",
            "range": 0.005},
        wd={"type": "l2", "scale": 0.},
        out_channel_num=10,
        activation={"type": "softmax"},
        bn={{ net_paras["bn"] }}))

    # Set up a sensor.
    # #########################################################################
    cifar_source = Cifar10FeedSource(
        name="CIFAR10",
        url='http://www.cs.toronto.edu/~kriz/cifar-10-binary.tar.gz',
        work_dir=AKID_DATA_PATH + '/cifar10',
        use_zca=True,
        num_train=50000,
        num_val=10000)

    sensor = FeedSensor(source_in=cifar_source,
                        batch_size=128,
                        name='data')

    # Summon a survivor.
    # #########################################################################
    survivor = Kid(
        sensor,
        brain,
        MomentumKongFu(base_lr={{ opt_paras["lr"] }},
                       momentum=0.5,
                       decay_rate=0.1,
                       decay_epoch_num=50),
        max_steps=200000,
        graph=graph,
    )

    survivor.setup()
    return survivor


if __name__ == "__main__":
    tune(setup, opt_paras_list, net_paras_list)



from __future__ import absolute_import
import tensorflow as tf


def compute_gradients(opt, v):
    return opt.compute_gradients(v)


def apply_gradients(opt, grad):
    return opt.apply_gradients(grad)


def MomentumOptimizer(lr, var_list=None, momentum=0.9, use_nesterov=False):
    return tf.train.MomentumOptimizer(lr, momentum, use_nesterov=use_nesterov)


def GradientDescentOptimizer(lr, var_list=None):
    return tf.train.GradientDescentOptimizer(lr)


def AdamOptimizer(lr, var_list=None):
    return tf.train.AdamOptimizer(lr)



from __future__ import absolute_import
from __future__ import print_function
import tensorflow as tf

from akid import AKID_DATA_PATH
from akid import MNISTFeedSource
from akid import FeedSensor
from akid.models import OneLayerBrain

# Flags for defining the tf.train.ClusterSpec
tf.app.flags.DEFINE_string("ps_hosts", "",
                           "Comma-separated list of hostname:port pairs")
tf.app.flags.DEFINE_string("worker_hosts", "",
                           "Comma-separated list of hostname:port pairs")

# Flags for defining the tf.train.Server
tf.app.flags.DEFINE_string("job_name", "", "One of 'ps', 'worker'")
tf.app.flags.DEFINE_integer("task_index", 0, "Index of task within the job")

FLAGS = tf.app.flags.FLAGS


def main(_):
  ps_hosts = FLAGS.ps_hosts.split(",")
  worker_hosts = FLAGS.worker_hosts.split(",")

  # Create a cluster from the parameter server and worker hosts.
  cluster = tf.train.ClusterSpec({"ps": ps_hosts, "worker": worker_hosts})

  # Create and start a server for the local task.
  server = tf.train.Server(cluster,
                           job_name=FLAGS.job_name,
                           task_index=FLAGS.task_index)

  if FLAGS.job_name == "ps":
    server.join()
  elif FLAGS.job_name == "worker":

    # Assigns ops to the local worker by default.
    with tf.device(tf.train.replica_device_setter(
        worker_device="/job:worker/task:%d" % FLAGS.task_index,
        cluster=cluster)):

      # Build model...
      source = MNISTFeedSource(name="MNIST",
                               url='http://yann.lecun.com/exdb/mnist/',
                               work_dir=AKID_DATA_PATH + '/mnist',
                               center=True,
                               scale=True,
                               num_train=50000,
                               num_val=10000)
      sensor = FeedSensor(name='data', source_in=source)

      sensor.forward()
      brain = OneLayerBrain(name="brain")
      # input = [sensor.data()]
      # input.extend(sensor.labels())
      input = [sensor.data(), sensor.labels()]
      brain.forward(input)
      loss = brain.loss
      global_step = tf.Variable(0)

      train_op = tf.train.AdagradOptimizer(0.01).minimize(
        loss, global_step=global_step)

      saver = tf.train.Saver()
      summary_op = tf.summary.merge_all()
      init_op = tf.global_variables_initializer()

    # Create a "supervisor", which oversees the training process.
    sv = tf.train.Supervisor(is_chief=(FLAGS.task_index == 0),
                             logdir="/tmp/train_logs",
                             init_op=init_op,
                             summary_op=None,
                             saver=saver,
                             global_step=global_step,
                             save_model_secs=600)

    # The supervisor takes care of session initialization, restoring from
    # a checkpoint, and closing when done or an error occurs.
    with sv.managed_session(server.target) as sess:
      # Loop until the supervisor shuts down or 1000000 steps have completed.
      step = 0
      while not sv.should_stop() and step < 1000000:
        # Run a training step asynchronously.
        # See `tf.train.SyncReplicasOptimizer` for additional details on how to
        # perform *synchronous* training.
        loss_value, _, step = sess.run([loss, train_op, global_step],
                           feed_dict=sensor.fill_feed_dict())
        print(loss_value)

    # Ask for all the services to stop.
    sv.stop()

if __name__ == "__main__":
  tf.app.run()



from __future__ import absolute_import
import tensorflow as tf

from akid.utils import glog as log
from akid.core.common import (
    TRAIN_SUMMARY_COLLECTION,
    VALID_SUMMARY_COLLECTION,
    TRAINING_DYNAMICS_COLLECTION,
)
from .. import computational_graph as general_cg
from . import computational_graph as cg


summary_writer = None
summary_ops = None


def init(dir=None):
    """
    Create the summary file writer.

    Args:
        dir: str
            The directory to save event files.
    """
    global summary_writer
    if not dir:
        dir = log.get_random_log_dir()
    summary_writer = tf.summary.FileWriter(dir)

    log.info("Summary event file will be saved to {}".format(dir))


def histogram(name, values, summary_on_grad=False, collections=None, step=None):
    """
    `step`, `summary_on_grad` are not used, just for compatibility with PyTorch.
    """
    tf.summary.histogram(name, values, collections)


def image(name, values, collections=None, step=None):
    tf.summary.image(name, values, collections=collections)


def scalar(name, value, collections=None, step=None):
    """
    Same as `histogram`.
    """
    if len(value.get_shape().as_list()) != 0:
        value = tf.reshape(value, [])
    tf.summary.scalar(name, value, collections=collections)


def add_graph():
    summary_writer.add_graph(tf.get_default_graph())


def add_scalar(name, value, step):
    summary = tf.Summary()
    summary.value.add(tag=name, simple_value=value)
    summary_writer.add_summary(summary, step)


def get_collection(name=None):
    if name:
        return tf.get_collection(name)
    else:
        ret = []
        ret.extend(tf.get_collection(TRAIN_SUMMARY_COLLECTION))
        ret.extend(tf.get_collection(VALID_SUMMARY_COLLECTION,))
        ret.extend(tf.get_collection(TRAINING_DYNAMICS_COLLECTION))

        return ret


def run_summary_op(op, feed_dict=None):
    summary_str = cg.sess.run(op, feed_dict=feed_dict)
    summary_writer.add_summary(summary_str, general_cg.get_step())


def merge(ops):
    return tf.summary.merge(ops)



"""
This module holds sources (feed source and tfrecord source) for CIFAR10
dataset. It supports using raw CIFAR10 images as input, as well as global
contrasted normalized, and ZCA whitened images (the common pre-processing done
on CIFAR dataset, starting from the Maxout paper).
"""


from __future__ import absolute_import
import os
import six.moves.cPickle as pickle

import numpy as np
import tensorflow as tf
from torchvision import datasets, transforms

from ..core.sources import (
    Source,
    InMemoryFeedSource,
    SupervisedSource,
    ClassificationTFSource,
    StaticSource
)
from .datasets import DataSet, DataSets
from six.moves import range

from akid import backend as A


class Cifar10Source(SupervisedSource):
    """
    A shared ancestor for different CIFAR10 sources. It is still an abstract
    class.
    """
    IMAGE_SIZE = 32
    SAMPLE_NUM = 50000

    def __init__(self, use_zca=False, **kwargs):
        """
        Args:
            use_zca: Boolean
                Use ZCA whitened data or not. If this is specified, the ZCA
                whitened data has to be in `work_dir` already.
        """
        super(Cifar10Source, self).__init__(**kwargs)
        self.use_zca = use_zca

    def _load_cifar10_python(self, filenames):
        """
        Load python version of Cifar10 dataset.
        """
        # Load the first batch of data to get shape info.
        filename = filenames[0]
        with open(filename, "rb") as f:
            tmp = pickle.load(f)
            data = tmp["data"]
            labels = np.array(tmp["labels"])

        # Load the rest.
        for filename in filenames[1:]:
            with open(filename, "rb") as f:
                tmp = pickle.load(f)
                data = np.append(data, tmp["data"], 1)
                labels = np.append(labels, tmp["labels"])

        data = np.reshape(data, [-1, 3, 32, 32])
        data = np.einsum("nchw->nhwc", data)

        return DataSet(data, labels)


class Cifar10FeedSource(Cifar10Source, InMemoryFeedSource):
    """
    A concrete `FeedSource` for Cifar10 dataset.
    """
    @property
    def shape(self):
        return [32, 32, 3]

    @property
    def label_shape(self):
        return [1]

    def _load(self):
        self._get_raw_data_if_not_yet()

        # Load training set into memory.
        train_filenames = [os.path.join(self.work_dir, 'cifar-10-batches-py',
                                        'data_batch_%d' % i)
                           for i in range(1, 6)]
        training_dataset = self._load_cifar10_python(train_filenames)

        test_filenames = [os.path.join(self.work_dir, 'cifar-10-batches-py',
                                       'test_batch')]
        test_dataset = self._load_cifar10_python(test_filenames)

        if self.use_zca:
            sample_num = len(train_filenames) * 10000
            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "train.npy"))
            imgs = imgs.reshape([sample_num, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)
            training_dataset = DataSet(
                imgs[0:self.num_train, ...],
                training_dataset.labels[0:self.num_train])

            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "test.npy"))
            imgs = imgs.reshape([10000, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)
            test_dataset = DataSet(imgs[0:self.num_val, ...],
                                   test_dataset.labels[0:self.num_val])

        return DataSets(training_dataset, test_dataset)


class Cifar10TFSource(Cifar10Source, ClassificationTFSource):
    """
    A concrete `Source` for Cifar10 dataset.
    """
    def _forward(self):
        """
        Construct input for CIFAR evaluation using the Reader ops.
        """
        if self.use_zca:
            self._read_from_tfrecord()
        else:
            self._read_from_bin()

    def _read_from_tfrecord(self):
        self._maybe_convert_to_tf()

        # Read and set up data tensors.
        filename = os.path.join(self.work_dir, 'cifar10_training.tfrecords')
        with tf.name_scope('input'):
            filename_queue = tf.train.string_input_producer([filename])
        self._training_datum, self._training_label \
            = self._get_sample_tensors_from_tfrecords(filename_queue)

        filename = os.path.join(self.work_dir, 'cifar10_test.tfrecords')
        with tf.name_scope('input'):
            filename_queue = tf.train.string_input_producer([filename])
        self._val_datum, self._val_label \
            = self._get_sample_tensors_from_tfrecords(filename_queue)

    def _maybe_convert_to_tf(self):
        """
        If tfrecords data are not available, convert it from the ZCA whitened
        data and downloaded labels.

        TODO: write the data pre-processing code so I won't need to rely on
        pylearn2.
        """
        TRAINING_TF_FILENAME = "cifar10_training"
        if not os.path.exists(
                os.path.join(self.work_dir,
                             TRAINING_TF_FILENAME + ".tfrecords")):
            # Read the numpy data in and convert it to TFRecord.
            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "train.npy"))
            imgs = imgs.reshape([Cifar10Source.SAMPLE_NUM, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)

            # Load training set into memory.
            train_filenames = [os.path.join(self.work_dir,
                                            'cifar-10-batches-py',
                                            'data_batch_%d' % i)
                               for i in range(1, 6)]
            training_labels = self._load_cifar10_python(train_filenames).labels
            self._convert_to_tf(imgs, training_labels, TRAINING_TF_FILENAME)

        TEST_TF_FILENAME = "cifar10_test"
        if not os.path.exists(
                os.path.join(self.work_dir, TEST_TF_FILENAME + ".tfrecords")):
            imgs = np.load(os.path.join(self.work_dir,
                                        "pylearn2_gcn_whitened",
                                        "test.npy"))
            imgs = imgs.reshape([10000, 3, 32, 32])
            imgs = np.einsum("nchw->nhwc", imgs)
            test_filenames = [os.path.join(self.work_dir,
                                           'cifar-10-batches-py',
                                           'test_batch')]
            test_labels = self._load_cifar10_python(test_filenames).labels
            self._convert_to_tf(imgs, test_labels, TEST_TF_FILENAME)

    def _get_sample_tensors_from_tfrecords(self, filename_queue):
        """
        Read from tfrecord file and return data tensors.

        Args:
            filename_queue: tf.train.string_input_producer
                A file name queue that gives string tensor for tfrecord names.

        Returns:
            (image, label): tuple of (rank-4 tf.float32 tensor and rank-1
                            tf.int32 tensor)
                individual sample that may be later put into a batch.
        """
        reader = tf.TFRecordReader()
        _, serialized_example = reader.read(filename_queue)
        features = tf.parse_single_example(
            serialized_example,
            # Defaults are not specified since both keys are required.
            features={
                'image_raw': tf.FixedLenFeature(
                    [Cifar10Source.IMAGE_SIZE, Cifar10Source.IMAGE_SIZE, 3],
                    tf.float32),
                'label': tf.FixedLenFeature([], tf.int64),
            })

        # Convert label from a scalar uint8 tensor to an int32 scalar.
        label = tf.cast(features['label'], tf.int32)
        image = features["image_raw"]

        return image, label

    def _read_from_bin(self):
        self._get_raw_data_if_not_yet()

        filenames = [os.path.join(self.work_dir, 'cifar-10-batches-bin',
                                  'data_batch_%d.bin' % i)
                     for i in range(1, 6)]
        datum, label = self._read_by_filenames(filenames)
        self._training_datum = datum
        self._training_label = label

        filenames = [os.path.join(self.work_dir, 'cifar-10-batches-bin',
                                  'test_batch.bin')]
        datum, label = self._read_by_filenames(filenames)
        self._val_datum = datum
        self._val_label = label

    def _read_by_filenames(self, filenames):
        for f in filenames:
            if not os.path.exists(f):
                raise ValueError('Failed to find file: ' + f)

        # Create a queue that produces the filenames to read.
        filename_queue = tf.train.string_input_producer(filenames)

        # Read examples from files in the filename queue.
        read_input = self._read_cifar10(filename_queue)
        reshaped_image = tf.cast(read_input.uint8image, tf.float32)

        return reshaped_image, read_input.label

    def _read_cifar10(self, filename_queue):
        """Reads and parses examples from CIFAR10 data files.

        Recommendation: if you want N-way read parallelism, call this function
        N times.  This will give you N independent Readers reading different
        files & positions within those files, which will give better mixing of
        examples.

        Args:
            filename_queue: A queue of strings with the filenames to read from.

        Returns:
            An object representing a single example, with the following fields:
            height: number of rows in the result (32)
            width: number of columns in the result (32)
            depth: number of color channels in the result (3)
            key: a scalar string Tensor describing the filename & record number
                for this example.
            label: an int32 Tensor with the label in the range 0..9.
            uint8image: a [height, width, depth] uint8 Tensor with the image
                data
        """

        class CIFAR10Record(object):
            pass
        result = CIFAR10Record()

        # Dimensions of the images in the CIFAR-10 dataset.
        # See http://www.cs.toronto.edu/~kriz/cifar.html for a description of
        # the input format.
        label_bytes = 1  # 2 for CIFAR-100
        result.height = 32
        result.width = 32
        result.depth = 3
        image_bytes = result.height * result.width * result.depth
        # Every record consists of a label followed by the image, with a
        # fixed number of bytes for each.
        record_bytes = label_bytes + image_bytes

        # Read a record, getting filenames from the filename_queue.  No
        # header or footer in the CIFAR-10 format, so we leave header_bytes
        # and footer_bytes at their default of 0.
        reader = tf.FixedLengthRecordReader(record_bytes=record_bytes)
        result.key, value = reader.read(filename_queue)

        # Convert from a string to a vector of uint8 that is record_bytes long.
        record_bytes = tf.decode_raw(value, tf.uint8)

        # The first bytes represent the label, which we convert from
        # uint8->int32.
        result.label = tf.cast(
            tf.slice(record_bytes, [0], [label_bytes]), tf.int32)

        # The remaining bytes after the label represent the image, which we
        # reshape from [depth * height * width] to [depth, height, width].
        depth_major = tf.reshape(
            tf.slice(record_bytes, [label_bytes], [image_bytes]),
            [result.depth, result.height, result.width])
        # Convert from [depth, height, width] to [height, width, depth].
        result.uint8image = tf.transpose(depth_major, [1, 2, 0])

        return result


class Cifar10TorchSource(StaticSource, SupervisedSource):
    def _setup(self):
        convert = transforms.Compose([
            transforms.ToTensor(),
            # lambda x: x.permute(2, 0, 1),
            transforms.Normalize([i/255 for i in [125.3, 123.0, 113.9]],
                                 [i/255 for i in [63.0, 62.1, 66.7]]),
        ])
        train_transform = transforms.Compose([
            transforms.RandomHorizontalFlip(),
            transforms.Pad(4),
            transforms.RandomCrop(32),
            convert,
        ])

        self.dataset = datasets.CIFAR10(self.work_dir, train=True, download=True,
                                        transform=train_transform)
        self.val_dataset = datasets.CIFAR10(self.work_dir, train=False,
                                            transform=convert)
    def _forward(self):
        pass


class Cifar10AkidSource(Source):
    def _setup(self):
        if not hasattr(self, "_data"):
            convert = transforms.Compose([
                transforms.ToTensor(),
                # lambda x: x.permute(2, 0, 1),
                transforms.Normalize([i/255 for i in [125.3, 123.0, 113.9]],
                                    [i/255 for i in [63.0, 62.1, 66.7]]),
            ])
            train_transform = transforms.Compose([
                transforms.RandomHorizontalFlip(),
                transforms.Pad(4),
                transforms.RandomCrop(32),
                convert,
            ])

            self._data = datasets.CIFAR10(self.work_dir, train=True, download=True,
                                            transform=train_transform)

    def set_mode(self, mode):
        super(Cifar10AkidSource, self).set_mode(mode)

        convert = transforms.Compose([
            transforms.ToTensor(),
            # lambda x: x.permute(2, 0, 1),
            transforms.Normalize([i/255 for i in [125.3, 123.0, 113.9]],
                                 [i/255 for i in [63.0, 62.1, 66.7]]),
        ])
        train_transform = transforms.Compose([
            transforms.RandomHorizontalFlip(),
            transforms.Pad(4),
            transforms.RandomCrop(32),
            convert,
        ])

        if mode == A.Mode.TRAIN:
            self._data = datasets.CIFAR10(self.work_dir, train=True, download=True,
                                          transform=train_transform)
        elif mode == A.Mode.VAL:
            self._data = datasets.CIFAR10(self.work_dir, train=False,
                                          transform=convert)
        else:
            raise ValueError("Mode {} not supported yet.".format(mode))

    @property
    def data(self):
        return self._data

    @property
    def size(self):
        return len(self._data)



from __future__ import absolute_import
import inspect

from tensorflow.python.ops.image_ops import _Check3DImage
from tensorflow.python.ops import math_ops
from tensorflow.python.framework import dtypes
import tensorflow as tf


def rescale_image(image):
    """Rescale `image` to range $[0, 1]$.

    It does this by dividing the largest value in the images.

    Args:
        image: 3-D tensor of shape `[height, width, channels]`.

    Returns:
        The rescaled image with same shape as `image`.

    Raises:
        ValueError: if the shape of 'image' is incompatible with this function.
    """
    _Check3DImage(image)

    image = math_ops.cast(image, dtype=dtypes.float32)
    image_max = tf.reduce_max(image)

    image = math_ops.div(image, image_max)
    return image


__all__ = [name for name, x in locals().items() if not inspect.ismodule(x)]



"""
A brain is the data processing engine to process data supplied by `Sensor` to
fulfill certain tasks. More specifically,

* it builds up blocks to form an arbitrary network
* offers sub-graphs for inference, loss, evaluation, summaries
* provides access to all data and parameters within

To use a brain, feed in data as a list, as how it is done in any other
blocks. Some pre-specified brains are available under `akid.models.brains`. An
example that sets up a brain using existing brains is::

    # ... first get a feed sensor
    sensor.forward()
    brain = OneLayerBrain(name="brain")
    input = [sensor.data(), sensor.labels()]
    brain.forward(input)

Note in this case, `data()` and `labels()` of `sensor` returns tensors. It is
not always the case. If it does not, saying return a list of tensors, you need
do things like::

    input = [sensor.data()]
    input.extend(sensor.labels())

Act accordingly.

Similarly, all blocks work this way.

A brain provides easy ways to connect blocks. For example, a one layer brain
can be built through the following::

    class OneLayerBrain(GraphBrain):
        def __init__(self, **kwargs):
            super(OneLayerBrain, self).__init__(**kwargs)
            self.attach(
                ConvolutionLayer(ksize=[5, 5],
                                strides=[1, 1, 1, 1],
                                padding="SAME",
                                out_channel_num=32,
                                name="conv1")
            )
            self.attach(ReLULayer(name="relu1"))
            self.attach(
                PoolingLayer(ksize=[1, 5, 5, 1],
                            strides=[1, 5, 5, 1],
                            padding="SAME",
                            name="pool1")
            )

            self.attach(InnerProductLayer(out_channel_num=10, name="ip1"))
            self.attach(SoftmaxWithLossLayer(
                class_num=10,
                inputs=[
                    {"name": "ip1", "idxs": [0]},
                    {"name": "system_in", "idxs": [1]}],
                name="loss"))

It assembles a convolution layer, a ReLU Layer, a pooling layer, an inner
product layer and a loss layer. To attach a block (layer) that directly takes
the outputs of the previous attached layer as inputs, just directly attach the
block. If `inputs` exists, the brain will fetch corresponding tensors by name
of the block attached and indices of the outputs of that layer. See the loss
layer above for an example. Note that even though there are multiple inputs for
the brain, the first attached layer of the brain will take the first of these
input by default, given the convention that the first tensor is the data, and
the remaining tensors are normally labels, which is not used till very late.
"""
from __future__ import absolute_import, division, print_function

import inspect

from .blocks import ProcessingLayer
from .systems import System, SequentialSystem, GraphSystem, SequentialGSystem
from .. import backend as A
from .common import (
    TRAIN_SUMMARY_COLLECTION,
    VALID_SUMMARY_COLLECTION,
)


class Brain(System, ProcessingLayer):
    """
    `Brain` supports variable sharing layer copy. By calling `get_copy` of the
    brain and call `forward` again, you will get a brain with the same
    configuration and that shares any variables original brain has.

    Note if `do_summary` and `moving_average_decay` are specified, it would
    override that option of any layers attached to this brain.
    """
    NAME = "Brain"

    def __init__(self, do_stat_on_norm=False, **kwargs):
        super(Brain, self).__init__(**kwargs)
        self.do_stat_on_norm = do_stat_on_norm

    def get_val_copy(self):
        """
        Get a copy for validation.
        """
        if A.backend() == A.TF:
            val_copy = self.get_copy()
        elif A.backend() == A.TORCH:
            val_copy = self
        else:
            raise ValueError("Not supported backend.")

        val_copy.set_val(True)

        return val_copy

    def get_shadow_copy(self):
        tower = self.get_copy()
        tower.set_shadow()

        return tower

    def set_val(self, val):
        """
        Change the state of the brain to validation.
        """
        assert type(val) is bool
        self.mode = A.Mode.VAL if val else A.Mode.TRAIN
        for b in self.blocks:
            b.set_val(val)

    def set_shadow(self):
        """
        Change the state of the brain to shadow replica.
        """
        super(Brain, self).set_shadow()
        for b in self.blocks:
            b.set_shadow()

    def switch_batch_monitoring_mode(self):
        """
        Switch on and off batch monitoring mode.
        """
        if self.is_mon:
            self.name = self.old_name
            self.set_flag("is_mon", False)
        else:
            # Save the old name and change status flag
            self.old_name = self.name
            self.name += "_mon"
            self.set_flag("is_mon", True)


    def set_do_summary_on_val_flag(self, v):
        super(Brain, self).set_do_summary_on_val_flag(v)
        for b in self.blocks:
            b.do_summary_on_val = v

    def get_filters(self, names=None):
        """
        A public interface to expose filters of this brain.

        Args:
            names: a list of str
                If not None, only get filters in that list.

        Returns:
            A list: Filters are returned as a list, ordered by the order they
                are added to the brain.
        """
        filter_list = []
        for b in self.blocks:
            if names:
                if b.name not in names:
                    continue
            for v in b.var_list:
                if v.requires_grad:
                    filter_list.append(v)

        return filter_list

    def num_parameters(self):
        total_parameters = 0
        for variable in self.get_filters():
            shape = A.get_shape(variable)
            variable_parametes = 1
            for dim in shape:
                variable_parametes *= dim
            total_parameters += variable_parametes

        return total_parameters

    def _pre_setup(self):
        super(Brain, self)._pre_setup()
        if self.is_val:
            self.log("Setting up val brain {} ...".format(self.name))
        else:
            self.log("Setting up brain {} ...".format(self.name))

    def _pre_forward(self, *args, **kwargs):
        super(Brain, self)._pre_forward(*args, **kwargs)
        self.log("Forwarding in {} mode.".format(self.mode), debug=True)

    def _post_forward(self, *args, **kwargs):
        self._gather_loss()
        self._gather_evals()
        self._gather_train_ops()

        super(Brain, self)._post_forward(*args, **kwargs)

    def _gather_loss(self):
        """
        Gather all losses in all blocks in this brain.
        """
        loss_list = []
        for b in self.blocks:
            if b.loss is not None:
                loss_list.append(b.loss)
        # The total loss is defined as the cross entropy loss plus all of the
        # weight decay terms (L2 loss).
        self._loss = A.add_n(loss_list, name='total_loss')

    def _gather_evals(self):
        """
        Gather all evaluation in all blocks in this brain.

        `self.eval` points to a list even there is only one evaluation metric.
        """
        eval_graph_list = []
        for b in self.blocks:
            if b.eval is not None:
                if type(b.eval) is list:
                    eval_graph_list.extend(b.eval)
                else:
                    eval_graph_list.append(b.eval)

        self._eval = eval_graph_list

        if self.is_val:
            verbose_eval_list = []
            for b in self.blocks:
                if b.verbose_eval is not None:
                    if type(b.verbose_eval) is list:
                        verbose_eval_list.extend(b.verbose_eval)
                    else:
                        verbose_eval_list.append(b.verbose_eval)

            if len(verbose_eval_list) > 0:
                self._verbose_eval = verbose_eval_list
            else:
                self._verbose_eval = None

    def _gather_train_ops(self):
        """
        Gather all train ops in all blocks in this brain.

        `self.train_op` points to a list.
        """
        # TODO: this method has a bad name, it is for some auxiliary ops that
        # should be run alongside with train_op in tensorflow.
        train_op_list = []
        for b in self.blocks:
            if b.train_op is not None:
                if type(b.train_op) is list:
                    train_op_list.extend(b.train_op)
                else:
                    train_op_list.append(b.train_op)

        self._train_op = train_op_list

    def on_para_update(self):
        ops = []
        for b in self.blocks:
            ops.extend(b.on_para_update())

        return ops

    def attach(self, *block_in):
        for b in block_in:
            t = type(b)
            if t is list or t is tuple:
                for l in b:
                    self.attach(l)
            else:
                # A brain should only contain data processing layers.
                assert issubclass(t, ProcessingLayer), \
                    "A `Brain` should only contain `ProcessingLayer`s."
                super(Brain, self).attach(b)
                # Pass options down.
                if self.moving_average_decay:
                    # Only pass it down when it is not None.
                    b.moving_average_decay = self.moving_average_decay
                b.do_stat_on_norm = self.do_stat_on_norm


class SequentialBrain(Brain, SequentialSystem):
    pass


class GraphBrain(Brain, GraphSystem):
    """
    A concrete class to build signal processing system (specifically neural
    network).

    To use the input data in a block using `inputs`, its name is
    "system_in". For example, say that the following layer uses the labels
    passed in::

        brain.attach(SoftmaxWithLossLayer(
            class_num=10,
            inputs=[{"name": "ip1", "idxs": [0]},
                    {"name": "system_in", "idxs": [1]}],
            name="loss"))
    """
    pass


class SeqentialGBrain(Brain, SequentialGSystem):
    pass


__all__ = [name for name, x in locals().items() if not inspect.ismodule(x)]
